'use strict';

/**
 * Wrapper for an HD graphics in pixel art games
 * The need for this entity comes from the fact that the HD graphics are downscaled to fit the
 * pixel art resolution. If an entity + huge sprite is downscaled, the children of this entity are downscaled as well resulting in 
 * awkward rescaling of children. Instead the developer passes the desired dimension in pixels and the sprite is
 * scaled down to this. Consider this component as a replace for sprites (it's actually a wrapper for sprites), but
 * note that this is actually an entity
 * @moduleName HDNineSlice
 * @snippet HDNineSlice|constructor
HDNineSlice({
    gamepixelDimension: new Rectangle(0, 0, ${1:32}, ${2:32}), // desired dimension in pixels
    imageName: '${3}',
    originRelative: new Vector2(${4:0.5}, ${5:0.5}),
})
 */

bento.define('components/hdnineslice', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/nineslice', 'bento/entity', 'bento/utils', 'bento/tween'], function (Bento, Vector2, Rectangle, Nineslice, Entity, Utils, Tween) {
    'use strict';

    return function (settings) {
        var name = settings.name || 'hdNineSlice';
        var position = settings.position;

        var gamepixelDimension = settings.gamepixelDimension || new Rectangle(0, 0, 0, 0);
        var viewportDimension = settings.viewportDimension || new Rectangle(0, 0, 0, 0);
        var inheritDimension = settings.inheritDimension || new Rectangle(0, 0, 0, 0);

        var scaleFactor = settings.scaleFactor || 0.144;
        var targetDimension = new Rectangle(0, 0, 0, 0);
        var nineslice = new Nineslice(settings);
        var parent;
        var behavior = {
            name: 'hdNinesliceBehavior',
            start: function start() {

                // apply dimensionType to targetDimension
                resize(gamepixelDimension, viewportDimension, inheritDimension);

                // hd graphic's parent dimension will become the desired dimension
                if (parent) {
                    parent.dimension.width = targetDimension.width;
                    parent.dimension.height = targetDimension.height;
                }

                nineslice.width = targetDimension.width / scaleFactor;
                nineslice.height = targetDimension.height / scaleFactor;
                entity.scale = new Vector2(scaleFactor, scaleFactor);

                if (entity.parent) {
                    entity.parent.dimension.x = -nineslice.width * nineslice.settings.originRelative.x * entity.scale.x;
                    entity.parent.dimension.y = -nineslice.height * nineslice.settings.originRelative.y * entity.scale.y;
                    var positioner = entity.parent.getComponent('positioner');
                    if (positioner) {
                        positioner.reposition();
                    }
                }
            },
            onParentAttached: function onParentAttached(data) {
                parent = data.parent;
            }
        };
        var entity = new Entity({
            name: name,
            position: position,
            components: [nineslice, behavior]
        });
        /**
         * Applies dimensionType to targetDimension
         */
        var resize = function resize(startdimension, addviewports, addparents) {
            var viewport = Bento.getViewport();
            var granddadDimension;

            targetDimension.width = startdimension.width;
            targetDimension.height = startdimension.height;

            targetDimension.width += addviewports.width * viewport.width;
            targetDimension.height += addviewports.height * viewport.height;

            if (typeof entity.parent.parent.dimension !== "undefined") {
                granddadDimension = entity.parent.parent.dimension;
                targetDimension.width += addparents.width * granddadDimension.width;
                targetDimension.height += addparents.height * granddadDimension.height;
            }
        };

        // convert gamepixelDimension to Rectangle
        if (!gamepixelDimension.isRectangle) {
            gamepixelDimension = new Rectangle(gamepixelDimension.x || 0, gamepixelDimension.y || 0, gamepixelDimension.width || 0, gamepixelDimension.height || 0);
        }

        return entity;
    };
});