'use strict';

/**
 * Wrapper for an HD graphics in pixel art games
 * The need for this entity comes from the fact that the HD graphics are downscaled to fit the
 * pixel art resolution. If an entity + huge sprite is downscaled, the children of this entity are downscaled as well resulting in 
 * awkward rescaling of children. Instead the developer passes the desired dimension in pixels and the sprite is
 * scaled down to this. Consider this component as a replace for sprites (it's actually a wrapper for sprites), but
 * note that this is actually an entity
 * @moduleName HDSprite
 * @snippet HDSprite|constructor
HDSprite({
    gamepixelDimension: new Rectangle(0, 0, ${1:32}, ${2:32}), // desired dimension in pixels
    imageName: '${3}',
    originRelative: new Vector2(${4:0.5}, ${5:0.5}),
    frameCountX: ${6:1},
    frameCountY: ${7:1},
    animations: {
        default: {
            speed: 0,
            frames: [0]
        }
    }
})
 */

bento.define('components/hdsprite', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/entity', 'bento/utils', 'bento/tween'], function (Bento, Vector2, Rectangle, Sprite, Entity, Utils, Tween) {
    'use strict';

    return function (settings) {
        var name = settings.name || 'hdSprite';
        var position = settings.position;

        var gamepixelDimension = settings.gamepixelDimension || new Rectangle(0, 0, 0, 0);
        var viewportDimension = settings.viewportDimension || new Rectangle(0, 0, 0, 0);
        var inheritDimension = settings.inheritDimension || new Rectangle(0, 0, 0, 0);

        var forceFitToParent = settings.forceFitToParent || false;

        var targetDimension = new Rectangle(0, 0, 0, 0);
        var hasWidth = !Utils.isEmpty(gamepixelDimension.width);
        var hasHeight = !Utils.isEmpty(gamepixelDimension.height);
        var sprite = new Sprite(settings);
        var parent;
        var behavior = {
            name: 'hdSpriteBehavior',
            start: function start() {
                var ratio;
                ratio = sprite.frameWidth / sprite.frameHeight;

                // apply dimensionType to targetDimension
                resize(gamepixelDimension, viewportDimension, inheritDimension);

                // hd graphic's parent dimension will become the desired dimension
                if (parent) {
                    parent.dimension.width = targetDimension.width;
                    parent.dimension.height = targetDimension.height;
                }

                entity.scale.x = targetDimension.width / sprite.frameWidth;
                entity.scale.y = targetDimension.height / sprite.frameHeight;

                // one of parameters was not given
                if (!hasWidth) {
                    entity.scale.x = entity.scale.y;
                    parent.dimension.width = targetDimension.height * ratio;
                }
                if (!hasHeight) {
                    entity.scale.y = entity.scale.x;
                    parent.dimension.height = targetDimension.width / ratio;
                }
                if (!hasWidth && !hasHeight) {
                    // reset
                    entity.scale.x = 1;
                    entity.scale.y = 1;
                    parent.dimension.width = sprite.frameWidth;
                    parent.dimension.height = sprite.frameHeight;
                }

                if (forceFitToParent && typeof entity.parent.parent.dimension !== "undefined") {
                    var granddadDimension = entity.parent.parent.dimension;

                    if (parent.dimension.height * parent.scale.y > granddadDimension.height) {
                        parent.scale.x = parent.scale.x * 1 / (parent.dimension.height / granddadDimension.height);
                        parent.scale.y = parent.scale.y * 1 / (parent.dimension.height / granddadDimension.height);
                    }
                    if (parent.dimension.width * parent.scale.x > granddadDimension.width) {
                        parent.scale.x = parent.scale.x * 1 / (parent.dimension.width / granddadDimension.width);
                        parent.scale.y = parent.scale.y * 1 / (parent.dimension.width / granddadDimension.width);
                    }
                }
                // update dimension x and y
                if (entity.parent) {
                    parent.dimension.x = -sprite.origin.x * entity.scale.x;
                    parent.dimension.y = -sprite.origin.y * entity.scale.y;
                    var positioner = entity.parent.getComponent('positioner');
                    if (positioner) {
                        positioner.reposition();
                    }
                }
            },
            onParentAttached: function onParentAttached(data) {
                parent = data.parent;
            }
        };
        var entity = new Entity({
            name: name,
            position: position,
            components: [sprite, behavior]
        });
        /**
         * Applies dimensionType to targetDimension
         */
        var resize = function resize(startdimension, addviewports, addparents) {
            var viewport = Bento.getViewport();
            var granddadDimension;

            // absolute
            if (hasWidth) {
                targetDimension.width = startdimension.width;
            }
            if (hasHeight) {
                targetDimension.height = startdimension.height;
            }

            targetDimension.width += addviewports.width * viewport.width;
            targetDimension.height += addviewports.height * viewport.height;

            if (typeof entity.parent.parent.dimension !== "undefined") {
                granddadDimension = entity.parent.parent.dimension;
                targetDimension.width += addparents.width * granddadDimension.width;
                targetDimension.height += addparents.height * granddadDimension.height;
            }

            //if ((dimensionType === 'relative' || dimensionType === 'parent') && entity && entity.parent && entity.parent.parent) {
            //    parentDimension = entity.parent.parent.dimension;
            //    // relative to grandparent [0...1]
            //    // note: not parent, because that's who the hdsprite is attached to, which presumably has no dimension yet
            //    if (hasWidth) {
            //        targetDimension.width = dimension.width * parentDimension.width;
            //    }
            //    if (hasHeight) {
            //        targetDimension.height = dimension.height * parentDimension.height;
            //    }
            //} else if (dimensionType === 'viewport') {
            //    // viewport [0...1]
            //    if (hasWidth) {
            //        targetDimension.width = dimension.width * viewport.width;
            //    }
            //    if (hasHeight) {
            //        targetDimension.height = dimension.height * viewport.height;
            //    }
            //} else {
            //    // absolute
            //    if (hasWidth) {
            //        targetDimension.width = dimension.width;
            //    }
            //    if (hasHeight) {
            //        targetDimension.height = dimension.height;
            //    }
            //}
        };

        // convert gamepixelDimension to Rectangle
        if (!gamepixelDimension.isRectangle) {
            gamepixelDimension = new Rectangle(gamepixelDimension.x || 0, gamepixelDimension.y || 0, gamepixelDimension.width || 0, gamepixelDimension.height || 0);
        }

        return entity;
    };
});