'use strict';

/**
 * Button with an icon on top
 * @moduleName IconButton
 * @snippet IconButton|snippet
IconButton({
    z: ${1:0},
    position: ${2:new Vector2(0, 0)},
    imageName: '${3:gui/buttons/button_yellow_wide}',
    frameCountX: ${4:1},
    frameCountY: ${5:3},

    // icon options
    icon: new Entity({}),
    iconSprite: new Sprite({}),
    iconName: 'gui/icons/',

    // offset of icon
    offset: new Vector2(0, 0),
    offsetWhenPressed: 0, // push pixels down when onButtonDown

    // icon transparency when inactive
    iconAlpha: 0.5,

    excited: true,
    onClick: function () {
        $6
    }
});
 */
bento.define('components/iconbutton', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/components/clickable', 'bento/entity', 'bento/eventsystem', 'bento/gui/clickbutton', 'bento/gui/counter', 'bento/gui/text', 'bento/utils', 'bento/tween', 'components/mask', 'bento/components/fill'], function (Bento, Vector2, Rectangle, Sprite, Clickable, Entity, EventSystem, ClickButton, Counter, Text, Utils, Tween, Mask, Fill) {
    'use strict';

    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        //var viewport = Bento.getViewport();
        var entitySettings = {};
        var entity;
        var iconEntity = settings.icon;
        var iconSprite = settings.iconSprite;
        var iconAlpha = settings.iconAlpha;
        var iconName = settings.iconName || '';
        var offset = settings.offset || new Vector2(0, 0);
        var offsetOnPressed = settings.offsetOnPressed || 0;
        var icon = iconEntity || new Entity({
            name: 'icon',
            components: [iconSprite || new Sprite({
                imageName: iconName,
                originRelative: new Vector2(0.5, 0.5)
            })]
        });
        var textDimension = new Rectangle((settings.textPos ? settings.textPos.x || 0 : 0) / (settings.scaleFactor || 1), (settings.textPos ? settings.textPos.y || 0 : 0) / (settings.scaleFactor || 1), (settings.textSize ? settings.textSize.width || 0 : 0) / (settings.scaleFactor || 1), (settings.textSize ? settings.textSize.height || 0 : 0) / (settings.scaleFactor || 1));
        textDimension.x -= textDimension.width / 2;
        textDimension.y -= textDimension.height / 2;
        var text;
        if (textDimension.width && textDimension.height) {
            if (Bento.isUITool) {
                text = new Fill({
                    name: settings.name + '_text',
                    color: [1, 0, 1, 0.5],
                    dimension: textDimension
                });
            } else {
                //I DON'T KNOW???
            }
        }

        var isExcited = settings.excited || false;
        var shineTween;
        var addExitedState = function addExitedState() {
            var shineBehavior = {
                name: 'shineBehavior',
                tween: function tween(data) {
                    shineTween = new Tween({
                        from: -entity.dimension.width / 2 - shine.dimension.width / 2,
                        to: entity.dimension.width / 2 + shine.dimension.width / 2,
                        in: 120,
                        ease: 'easeInOutCubic',
                        onUpdate: function onUpdate(v, t) {
                            shine.position.x = v;
                        },
                        onComplete: shineBehavior.tween
                    });
                },
                start: function start(data) {
                    shineBehavior.tween(data);
                },
                destroy: function destroy() {
                    if (shineTween) {
                        shineTween.removeSelf();
                    }
                },
                update: function update() {
                    shine.visible = isExcited;
                    shineOverlay.visible = isExcited;
                }
            };
            var shineSprite = new Sprite({
                imageName: settings.shineImage || 'gui/buttons/shine',
                originRelative: new Vector2(0.5, 0.5)
            });
            var shineMask = new Mask({
                // reference to rectangle, coordinates relative to parent's space
                rectangle: entity.dimension.clone().grow(-48),
                // reference to sprite
                sprite: shineSprite
            });
            var shine = new Entity({
                name: 'shine',
                position: new Vector2(0, 0),
                components: [shineSprite, shineMask, shineBehavior]
            });
            var shineOverlay = new Entity({
                name: 'shineOverlay',
                position: new Vector2(0, 0),
                components: [new Sprite({
                    imageName: settings.imageName + '-excite',
                    originRelative: new Vector2(0.5, 0.5)
                })]
            });

            entity.attach(shine);
            entity.attach(shineOverlay);
            // put them directly over the sprite (otherwise goes over the icon as well)
            entity.moveComponentTo(shine, 1);
            entity.moveComponentTo(shineOverlay, 2);
        };

        // setup icon position
        icon.position = new Vector2(offset.x / (settings.scaleFactor || 1), offset.y / (settings.scaleFactor || 1));
        // defaults
        entitySettings.frameCountX = entitySettings.frameCountX || 1;
        entitySettings.frameCountY = entitySettings.frameCountY || 3;

        // take over properties
        Utils.extend(entitySettings, settings);

        // overwrite callbacks
        entitySettings.onButtonDown = function () {
            icon.position.y = offset.y + offsetOnPressed;
            if (settings.onButtonDown) {
                settings.onButtonDown.apply(entity);
            }
        };
        entitySettings.onButtonUp = function () {
            icon.position.y = offset.y;
            if (settings.onButtonUp) {
                settings.onButtonUp.apply(entity);
            }
        };

        // merge components array
        entitySettings.components = [icon].concat(settings.components || []);

        if (text) {
            entitySettings.components.unshift(text);
        }

        if (settings.scaleFactor) {
            entitySettings.scale = new Vector2(settings.scaleFactor, settings.scaleFactor);
        }

        if (!entitySettings.scale) {
            // default scale
            entitySettings.scale = new Vector2(0.144, 0.144);
        }

        // spawn entity
        entity = new ClickButton(entitySettings);

        //set parent's dimensions to match
        entity.attach({
            name: 'iconbuttonBehavior',
            start: function start() {
                if (entity.parent && entity.parent.dimension.width === 0 && entity.parent.dimension.height === 0) {
                    entity.parent.dimension.width = entity.dimension.width * entitySettings.scale.x;
                    entity.parent.dimension.height = entity.dimension.height * entitySettings.scale.y;
                    entity.parent.dimension.x = -entity.dimension.width * 0.5 * entity.scale.x;
                    entity.parent.dimension.y = -entity.dimension.height * 0.5 * entity.scale.y;
                    var positioner = entity.parent.getComponent('positioner');
                    if (positioner) {
                        positioner.reposition();
                    }
                }
            }
        });

        // allow to change icon
        entity.extend({
            changeIcon: function changeIcon(newIcon, offsetPosition) {
                icon.getComponent('sprite', function (sprite) {
                    sprite.setup(newIcon);
                    if (offsetPosition) {
                        icon.position = offsetPosition.clone();
                    }
                });
            }
        });

        if (iconAlpha) {
            // add behavior to set alpha of icon when button is disabled
            entity.attach({
                name: 'alpha',
                update: function update() {
                    if (entity.isActive()) {
                        icon.alpha = 1;
                    } else {
                        icon.alpha = iconAlpha;
                    }
                }
            });
        }

        // SPECIFIC TO BUSTED
        if (Bento.assets.hasAsset(settings.imageName + '-excite', 'images')) {
            addExitedState();

            Object.defineProperty(entity, 'excited', {
                get: function get() {
                    return isExcited;
                },
                set: function set(value) {
                    isExcited = value;
                }
            });
        }

        return entity;
    };
});