'use strict';

/**
 * Attach this to a ScrollingList's item to mask it
 * @moduleName ScrollingMask
 * @snippet ScrollingMask|constructor
ScrollingMask({
    boundingBox: new Rectangle(-16, -16, 32, 32),
    direction: 'x',
    scrollingList: scrollingList
})
 */
bento.define('components/scrollingmask', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/entity', 'bento/eventsystem', 'bento/utils', 'bento/tween', 'components/mask'], function (Bento, Vector2, Rectangle, Entity, EventSystem, Utils, Tween, Mask) {
    'use strict';
    // dependency of ScrollingMask

    var MaskBehavior = function MaskBehavior(settings) {
        var direction = settings.direction || 'x';
        var boundingBox = settings.boundingBox.clone();
        var original = {
            x: boundingBox.x,
            y: boundingBox.y
        };
        var sharpness = settings.isText ? 4 : 1;
        var ancestors = settings.ancestors;
        var mask = new Mask({
            sprite: settings.sprite,
            rectangle: boundingBox
        });
        var entity;
        return {
            name: 'maskingBehavior',
            start: function start() {
                boundingBox.width *= sharpness;
                boundingBox.height *= sharpness;
            },
            attached: function attached(data) {
                entity = data.entity;
                entity.attach(mask);
            },
            removed: function removed() {
                entity.remove(mask);
            },
            update: function update() {
                var scroll = settings.scrollingList.getScroll();
                // sets the mask's boundingbox
                boundingBox[direction] = (original[direction] - scroll) * sharpness;
                // offset with each parent's position
                Utils.forEach(ancestors, function (parent, i, l, breakLoop) {
                    boundingBox[direction] -= parent.position[direction] * sharpness;
                });
            }
        };
    };
    var ScrollingMask = function ScrollingMask(settings) {
        var entity;
        var direction = settings.direction || 'x'; // scroll direction
        var boundingBox = settings.boundingBox;
        var scrollingList = settings.scrollingList;
        var loopComponents = function loopComponents(parent, parents) {
            var ancestors = parents.concat(parent);
            Utils.forEach(parent.components, function (child, i, l, breakLoop) {
                if (child.name === 'sprite') {
                    // mask this sprite by attaching extra behavior
                    entity.attach(new MaskBehavior({
                        direction: direction,
                        boundingBox: boundingBox,
                        sprite: child,
                        ancestors: ancestors,
                        scrollingList: scrollingList
                    }));
                } else if (child.name === 'text' || child.name === 'shadowText') {
                    // note: this would not be needed if scale was properly taken into account(?)
                    entity.attach(new MaskBehavior({
                        direction: direction,
                        boundingBox: boundingBox,
                        sprite: child.getComponent('sharpnessScaler').getComponent('sprite'),
                        ancestors: ancestors.concat([child]),
                        scrollingList: scrollingList,
                        isText: true // takes default sharpness into account
                    }));
                } else if (child.isEntity) {
                    // go deeper
                    loopComponents(child, ancestors);
                }
            });
        };
        var component = {
            name: 'scrollingMask',
            start: function start() {
                // loop through components
                loopComponents(entity, []);
            },
            attached: function attached(data) {
                entity = data.entity;
            }
        };
        return component;
    };
    return ScrollingMask;
});