'use strict';

bento.define('entities/hintarea', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/entity', 'bento/eventsystem', 'bento/gui/text', 'bento/utils', 'bento/screen', 'bento/tween', 'globals', 'bento/components/eventlistener', 'modules/box2d', 'components/pathfollower'], function (Bento, Vector2, Rectangle, Sprite, Entity, EventSystem, Text, Utils, Screen, Tween, Globals, EventListener, Box2D, PathFollower) {

    var decomp = window.decomp;
    var earcut = window.earcut;
    var PIXI = window.PIXI;

    var terrainFilter = new Box2D.Filter(Globals.Categories.WALL | Globals.Categories.SLICEABLE);
    var wallFilter = new Box2D.Filter(Globals.Categories.WALL);

    // old colors for bento renderer
    var terrainColor = Utils.color('#f1892d');
    var wallColor = Utils.color('#000000');
    var seamColor = Utils.color('#66442218');

    var Terrain = function Terrain(settings) {
        var tiled = settings.tiled,
            properties = settings.properties,
            position = settings.position,
            rotation = settings.rotation,
            fixture = settings.fixture,
            fixtures = settings.fixtures,
            _settings$type = settings.type,
            type = _settings$type === undefined ? 'static' : _settings$type,
            _settings$isWall = settings.isWall,
            isWall = _settings$isWall === undefined ? false : _settings$isWall,
            _settings$autoType = settings.autoType,
            autoType = _settings$autoType === undefined ? false : _settings$autoType,
            _settings$friction = settings.friction,
            friction = _settings$friction === undefined ? 0.2 : _settings$friction,
            _settings$restitution = settings.restitution,
            restitution = _settings$restitution === undefined ? 0 : _settings$restitution,
            _settings$density = settings.density,
            density = _settings$density === undefined ? 1.0 : _settings$density,
            _settings$follow = settings.follow,
            follow = _settings$follow === undefined ? null : _settings$follow;


        var concavePoly = void 0;

        var edgeInfo = {};
        var outerEdges = [];

        var pixiGraphic = new PIXI.Graphics();
        var pixiPolygon = void 0;
        var outerPolygon = void 0; //temp
        var pixiColor = void 0;

        var filter = isWall ? wallFilter : terrainFilter;
        var fillColor = isWall ? 0x5a6988 : 0xfeae34;
        var lineColor = isWall ? 0x3a4466 : 0xf77622;

        var getEdgeKey = function getEdgeKey(v1, v2) {
            // sort v1,v2 so key will always be the same regardless of direction
            if (v2.y < v1.y) {
                var tmp = v1;
                v1 = v2;
                v2 = tmp;
            }
            if (v2.x < v1.x) {
                var _tmp = v1;
                v1 = v2;
                v2 = _tmp;
            }
            // key should be reasonably short and not contain floating point innacuracies
            var x1 = Utils.round(v1.x, 0.01);
            var y1 = Utils.round(v1.y, 0.01);
            var x2 = Utils.round(v2.x, 0.01);
            var y2 = Utils.round(v2.y, 0.01);

            return x1 + ',' + y1 + ':' + x2 + ',' + y2;
        };

        var identifyEdges = function identifyEdges() {
            edgeInfo = {};
            outerEdges = [];
            body.fixtures.forEach(function (f) {
                var vertices = f.shape.GetVertices().map(Box2D.toBento);
                Utils.forEachEdge(vertices, function (v1, v2) {
                    var key = getEdgeKey(v1, v2);
                    var edge = edgeInfo[key];
                    if (!edge) {
                        edge = { v1: v1, v2: v2, count: 0 };
                        edgeInfo[key] = edge;
                    }
                    edge.count++;
                });
            });
            for (var key in edgeInfo) {
                var edge = edgeInfo[key];
                if (edge.count === 1) {
                    outerEdges.push(edge);
                }
            }
            buildOuterPolygon();
        };
        var isOuterEdge = function isOuterEdge(v1, v2) {
            var key = v1.x + ',' + v1.y + ':' + v2.x + ',' + v2.y;
            var edge = edgeInfo[key];
            return edge && edge.count === 1;
        };

        /**
         * Helper function for slicing.
         * Returns true if the given point lies on an outer edge.
         * (inner edges are those that are shared by two shapes in this body)
         */
        var onOuterEdge = function onOuterEdge(p) {
            for (var i = 0; i < outerEdges.length; i++) {
                if (Utils.pointOnEdge(p, outerEdges[i].v1, outerEdges[i].v2)) {
                    return true;
                }
            }
            return false;
        };

        var buildOuterPolygon = function buildOuterPolygon() {
            var edges = outerEdges.slice();
            for (var i = 0; i < edges.length - 1; i++) {
                var edge = edges[i];
                for (var j = i + 1; j < edges.length; j++) {
                    var nextEdge = edges[j];
                    if (Utils.approxEqualVec(edge.v2, nextEdge.v1, 0.001)) {
                        edges[j] = edges[i + 1];
                        edges[i + 1] = nextEdge;
                        break;
                    } else if (Utils.approxEqualVec(edge.v2, nextEdge.v2, 0.001)) {
                        var _ref = [nextEdge.v2, nextEdge.v1];
                        nextEdge.v1 = _ref[0];
                        nextEdge.v2 = _ref[1];

                        edges[j] = edges[i + 1];
                        edges[i + 1] = nextEdge;
                        break;
                    }
                }
                // this error isn't true? In fact breaking here causes incorrect behavior.
                // Utils.log(`Couldn't find next node. Body has two disconnected shapes??`);
                // break;
            }
            outerPolygon = [outerEdges[0].v1];
            edges.forEach(function (edge) {
                outerPolygon.push(edge.v2);
            });

            // update pixi graphics
            pixiPolygon = new PIXI.Polygon(outerPolygon.map(function (v) {
                return new PIXI.Point(v.x, v.y);
            }));
            pixiGraphic = new PIXI.Graphics();

            // This approach should work for masking... but it doesn't!
            /*        
            let filledPolygon = new PIXI.Graphics();
            let texture = Utils.newTexture(isWall ? 'texture-wall' : 'texture-cheese');
            let sprite = new PIXI.Sprite(texture);
            sprite.anchor.x = 0.5;
            sprite.anchor.y = 0.5;
            filledPolygon.clear();
            // pixiGraphic.lineStyle(1, lineColor);
            filledPolygon.beginFill(fillColor);
            filledPolygon.drawPolygon(pixiPolygon);
            filledPolygon.endFill();
             sprite.mask = filledPolygon;
            pixiGraphic.addChild(sprite);
            pixiGraphic.addChild(filledPolygon);
            */

            // WIP mesh approach
            /*
            let texture = Utils.newTexture(isWall ? 'texture-wall' : 'texture-cheese');
            let plane = new PIXI.mesh.Plane(texture, pixiPolygon);
            let graphics = new PIXI.Graphics();
            pixiGraphic.addChild(plane);
            pixiGraphic.addChild(graphics);
            graphics.drawPolygon(pixiPolygon);
            */
            pixiGraphic.clear();
            pixiGraphic.lineStyle(1, lineColor);
            pixiGraphic.beginFill(fillColor);
            pixiGraphic.drawPolygon(pixiPolygon);
            pixiGraphic.endFill();

            // console.log(Bento.getRenderer().getPixiRenderer());
        };

        var onBeginContact = function onBeginContact() {};

        if (tiled) {

            position = new Vector2(tiled.x, tiled.y);
            rotation = tiled.rotation;

            if (tiled.polygon) {
                concavePoly = tiled.polygon.map(function (v) {
                    return [v.x, v.y];
                });
                decomp.makeCCW(concavePoly);
                var convexPolys = decomp.quickDecomp(concavePoly);
                Utils.triangulateBadPolys(convexPolys);

                fixtures = convexPolys.map(function (poly) {
                    return {
                        filter: filter,
                        shape: new Box2D.PolygonShape(poly.map(function (v) {
                            return new Vector2(v[0], v[1]);
                        })),
                        isSensor: true,
                        onBeginContact: onBeginContact
                    };
                });
            } else {
                // assume box shape
                var vertices = [new Vector2(tiled.width, 0), new Vector2(tiled.width, tiled.height), new Vector2(0, tiled.height), new Vector2(0, 0)];
                fixture = {
                    filter: filter,
                    shape: new Box2D.PolygonShape(vertices),
                    friction: friction,
                    isSensor: true,
                    onBeginContact: onBeginContact
                };
            }
        }

        var body = new Box2D.Body({
            type: type,
            fixtures: fixtures
        });

        var behavior = {
            name: 'behavior',
            start: function start() {},
            draw: function draw(_ref2) {
                var renderer = _ref2.renderer;

                fixtures.forEach(function (f) {
                    Box2D.drawShape(renderer, f.shape);
                });
            }
        };

        var entity = new Entity({
            z: isWall ? Globals.Layers.WALL : Globals.Layers.TERRAIN,
            name: 'hintArea',
            family: ['hintArea'],
            position: position,
            rotation: rotation,
            components: [behavior, body]
        });

        return entity;
    };

    Terrain.color = terrainColor;
    Terrain.filter = terrainFilter;

    return Terrain;
});