'use strict';

bento.define('entities/lemming', ['bento', 'bento/components/sprite', 'bento/math/vector2', 'bento/math/rectangle', 'bento/entity', 'bento/eventsystem', 'bento/components/eventlistener', 'bento/utils', 'bento/tween', 'globals', 'modules/box2d', 'modules/savemanager', 'modules/skinmanager'], function (Bento, Sprite, Vector2, Rectangle, Entity, EventSystem, EventListener, Utils, Tween, Globals, Box2D, SaveManager, SkinManager) {

    var solidOnly = new Box2D.Filter(Globals.Categories.DEFAULT, Globals.Categories.WALL | Globals.Categories.SLICEABLE);

    /**
     * Returns a table of sprite settings for the current character skin.
     */
    var getAnimations = function getAnimations(characterId) {
        if (!characterId) {
            characterId = SaveManager.load('character');
        }
        var walkAnimSpeed = 0.125 * Globals.simulationSpeed;
        return {
            'idle': {
                imageName: 'character' + characterId + '-idle',
                originRelative: new Vector2(0.5, 0.8),
                frameCountX: 5,
                animations: {
                    default: { speed: 0.125, frames: [0, 1, 2, 3, 4] }
                }
            },
            'walk': {
                imageName: 'character' + characterId + '-walk',
                originRelative: new Vector2(0.5, 0.8),
                frameCountX: 8,
                animations: {
                    default: { speed: walkAnimSpeed, frames: [0, 1, 2, 3, 4, 5, 6, 7] }
                }
            },
            'victory': {
                imageName: 'character' + characterId + '-victory',
                originRelative: new Vector2(0.5, 0.8),
                frameCountX: 6,
                animations: {
                    default: { speed: 0.125, frames: [0, 1, 2, 3, 4, 5] }
                }
            }
        };
    };

    /**
     * The player character!
     */
    var Lemming = function Lemming(settings) {
        var position = settings.position,
            _settings$facing = settings.facing,
            facing = _settings$facing === undefined ? 'right' : _settings$facing,
            characterId = settings.characterId,
            randomizeCharacterId = settings.randomizeCharacterId,
            _settings$isRescued = settings.isRescued,
            isRescued = _settings$isRescued === undefined ? false : _settings$isRescued;


        if (!characterId) {
            if (randomizeCharacterId) {
                characterId = SkinManager.getRandomUnlockedCharacter();
                if (characterId === SaveManager.load('character')) {
                    // reduce chance of getting the main character
                    characterId = SkinManager.getRandomUnlockedCharacter();
                }
            } else {
                characterId = SaveManager.load('character');
            }
        }

        position.y -= 5;

        var mode = 'idle'; // idle, walk, victory 
        var turnCooldown = 0;
        var prevPosition = position.clone();

        // a kind of running average.
        // If this drops below a threshold we can assume the lemming is stuck against a wall.
        // Otherwise there may be movable rubble in the way.
        var averageDistancePerStep = new Vector2(0, 0);

        var anims = getAnimations(characterId);
        var sprite = new Sprite(anims[mode]);

        var firstSliceListener = new EventListener({
            eventName: 'firstSlice',
            onEvent: function onEvent() {
                sprite.setup(anims['walk']);
                mode = 'walk';
                turnCooldown = 10;
            }
        });

        // begin victory animation when we reach the goal or pull the lever.
        var hasWon = false;
        var doVictoryAnim = function doVictoryAnim() {
            if (!hasWon) {
                hasWon = true;
                sprite.setup(anims['victory']);
                mode = 'victory';
                if (Globals.cageCharacter) {
                    // turn to face the first lemming that's not me
                    var lemmings = Bento.objects.getByFamily('lemming').concat(Bento.objects.getByFamily('cage'));
                    for (var i = 0; i < lemmings.length; i++) {
                        var e = lemmings[i];
                        if (e !== entity) {
                            facing = e.position.x > entity.position.x ? 'right' : 'left';
                            break;
                        }
                    }
                }
            }
        };
        var winListener = new EventListener({
            eventName: 'win',
            onEvent: doVictoryAnim
        });
        var pullLeverListener = new EventListener({
            eventName: 'pullLever',
            onEvent: doVictoryAnim
        });

        var bodyFixture = {
            filter: new Box2D.Filter(Globals.Categories.PLAYER, Globals.Categories.WALL | Globals.Categories.GOAL | Globals.Categories.COLLECTIBLE | Globals.Categories.REGION),
            shape: new Box2D.CircleShape(5),
            friction: 0.1,
            restitution: 0.1,
            density: 1.0,
            onPostSolve: function onPostSolve(other, contact, impulse) {
                if (mode !== 'walk' || turnCooldown > 0 || averageDistancePerStep.sqrMagnitude() > 0.0005 * Bento.getGameSpeed()) {
                    return;
                }
                facing = facing === 'left' ? 'right' : 'left';
                turnCooldown = 10;

                if (Globals.failOnTurn) {
                    EventSystem.fire('fail');
                }
            }
        };
        var smallerBodyFixture = {
            filter: new Box2D.Filter(Globals.Categories.PLAYER, Globals.Categories.SLICEABLE),
            shape: new Box2D.CircleShape(4.5),
            friction: 0.1,
            restitution: 0.1,
            density: 0.0,
            onPostSolve: bodyFixture.onPostSolve
        };

        var inactiveColor = Utils.color('#ff0000');
        var activeColor = Utils.color('#ffff00');

        var floorSensor = {
            shape: new Box2D.CircleShape(3, new Vector2(0, 4)),
            filter: solidOnly,
            isSensor: true,
            contacts: 0,
            onBeginContact: function onBeginContact(other, contact) {
                this.contacts++;
            },
            onEndContact: function onEndContact(other, contact) {
                this.contacts--;
            }
        };

        // let leftSensor = {
        //     shape: new Box2D.CircleShape(3, new Vector2(-3, -2)),
        //     filter: solidOnly,
        //     isSensor: true,
        //     touching: false,
        //     onBeginContact: function (other, contact) {
        //         facing = 'right';
        //         this.touching = true;
        //     },
        //     onEndContact: function (other, contact) {
        //         this.touching = false;
        //     },
        // };
        // let rightSensor = {
        //     shape: new Box2D.CircleShape(3, new Vector2(3, -2)),
        //     filter: solidOnly,
        //     isSensor: true,
        //     touching: false,
        //     onBeginContact: function (other, contact) {
        //         facing = 'left';
        //         this.touching = true;
        //     },
        //     onEndContact: function (other, contact) {
        //         this.touching = false;
        //     },
        // };

        /**
         * Like update, but tied to Box2D step so it may run multiple times in one frame.
         */
        var onStep = function onStep() {
            var distance = position.subtract(prevPosition);
            averageDistancePerStep.addTo(distance);
            averageDistancePerStep.scale(0.8);

            // if (entity.ticker % 10 === 0) {
            //     console.log(averageDistancePerStep.sqrMagnitude());
            // }

            switch (mode) {
                case 'idle':
                    // zzz
                    break;
                case 'walk':
                    var vel = body.getLinearVelocity();
                    if (facing === 'left') {
                        body.setLinearVelocity(new Vector2(-30, vel.y));
                    } else {
                        body.setLinearVelocity(new Vector2(+30, vel.y));
                    }
                    if (turnCooldown > 0) {
                        turnCooldown -= 1.0;
                    }
                    body.setAwake(true);
                    break;
                case 'victory':
                    break;
            }

            prevPosition.x = position.x;
            prevPosition.y = position.y;
        };

        var body = new Box2D.Body({
            type: 'dynamic',
            fixedRotation: true,
            onStep: onStep,
            fixtures: [bodyFixture, smallerBodyFixture, floorSensor]
        });

        var frame = 0.0,
            oldFrame = 0.0;

        var behavior = {
            name: 'behavior',
            start: function start() {
                if (isRescued) {
                    doVictoryAnim();
                    body.applyImpulse(new Vector2(0, -6), position);
                }
            },
            update: function update(data) {
                switch (mode) {
                    case 'idle':
                        break;
                    case 'walk':
                        break;
                    case 'victory':
                        oldFrame = frame;
                        frame = sprite.getCurrentFrame();
                        if (frame >= 2.0 && oldFrame < 2.0 && floorSensor.contacts > 0) {
                            body.applyImpulse(new Vector2(0, -4), position);
                        }
                        break;
                }
            },
            draw: function draw(_ref) {
                var renderer = _ref.renderer;

                if (facing === 'left') {
                    // flip the sprite
                    renderer.scale(-1, 1);
                }
            }
        };

        // let debugDraw = {
        //     name: 'debugDraw',
        //     draw ({renderer}) {
        //         Box2D.drawShape(renderer, bodyFixture.shape);
        //         Box2D.drawShape(renderer, floorSensor.shape, floorSensor.contacts ? activeColor : inactiveColor);
        //     }
        // };

        var entity = new Entity({
            z: Globals.Layers.LEMMING,
            name: 'lemming',
            family: ['lemming'],
            position: position,
            scale: new Vector2(0.8, 0.8),
            components: [body, behavior, sprite, firstSliceListener, winListener, pullLeverListener]
        });

        return entity;
    };

    Lemming.getAnimations = getAnimations;

    return Lemming;
});