'use strict';

bento.define('entities/terrain', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/entity', 'bento/eventsystem', 'bento/gui/text', 'bento/utils', 'bento/screen', 'bento/tween', 'globals', 'bento/components/eventlistener', 'modules/box2d', 'components/pathfollower', 'modules/tapticengine', 'modules/savemanager'], function (Bento, Vector2, Rectangle, Sprite, Entity, EventSystem, Text, Utils, Screen, Tween, Globals, EventListener, Box2D, PathFollower, TapticEngine, SaveManager) {
    var _slicedToArray = function () {
        function sliceIterator(arr, i) {
            var _arr = [];
            var _n = true;
            var _d = false;
            var _e = undefined;

            try {
                for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
                    _arr.push(_s.value);

                    if (i && _arr.length === i) break;
                }
            } catch (err) {
                _d = true;
                _e = err;
            } finally {
                try {
                    if (!_n && _i["return"]) _i["return"]();
                } finally {
                    if (_d) throw _e;
                }
            }

            return _arr;
        }

        return function (arr, i) {
            if (Array.isArray(arr)) {
                return arr;
            } else if (Symbol.iterator in Object(arr)) {
                return sliceIterator(arr, i);
            } else {
                throw new TypeError("Invalid attempt to destructure non-iterable instance");
            }
        };
    }();

    var getThemeInfo = function getThemeInfo() {
        var theme = Globals.theme || SaveManager.load('theme');

        if (theme === 'wave-blue') {
            return {
                textureName: 'texture/solid-wide',
                scale: 0.24,
                lineColor: 0x3f4dd1
            };
        }

        var _theme$split = theme.split('-'),
            _theme$split2 = _slicedToArray(_theme$split, 2),
            shape = _theme$split2[0],
            color = _theme$split2[1];

        var angle = 0;
        var scale = 0.1;
        var lineColor = 0x3f4dd1;
        switch (shape) {
            case 'wave':
                scale = 0.10;break;
            case 'diagonal':
                scale = 0.08;angle = Utils.toRadian(30);break;
            case 'stone':
                scale = 0.14;break;
            case 'spot':
                scale = 0.20;break;
            case 'spiky':
                scale = 0.10;break;
            case 'ring':
                scale = 0.12;break;
            case 'stars':
                scale = 0.12;break;
        }
        switch (color) {
            case 'blue':
                lineColor = 0x3f4dd1;break;
            case 'brown':
                lineColor = 0x884942;break;
            case 'dark':
                lineColor = 0x191a20;break;
            case 'green':
                lineColor = 0x217e59;break;
            case 'orange':
                lineColor = 0xdb532c;break;
            case 'pink':
                lineColor = 0xb73278;break;
            case 'purple':
                lineColor = 0x6e34a5;break;
        }
        return {
            textureName: 'texture/solid-' + theme,
            angle: angle,
            scale: scale,
            lineColor: lineColor
        };
    };

    var decomp = window.decomp;
    var earcut = window.earcut;
    var PIXI = window.PIXI;

    var terrainFilter = new Box2D.Filter(Globals.Categories.SLICEABLE);
    var wallFilter = new Box2D.Filter(Globals.Categories.WALL);

    var nextAllowedHapticTime = 0;

    var globalEdges = {};

    var clearGlobalEdges = function clearGlobalEdges() {
        globalEdges = {};
    };
    var addGlobalEdge = function addGlobalEdge(v1, v2) {
        var key = Utils.getEdgeKey(v1, v2);
        var count = globalEdges[key] || 0;
        globalEdges[key] = count + 1;
    };

    var Terrain = function Terrain(settings) {
        var name = settings.name,
            tiled = settings.tiled,
            properties = settings.properties,
            position = settings.position,
            rotation = settings.rotation,
            fixture = settings.fixture,
            fixtures = settings.fixtures,
            _settings$type = settings.type,
            type = _settings$type === undefined ? 'static' : _settings$type,
            _settings$isWall = settings.isWall,
            isWall = _settings$isWall === undefined ? false : _settings$isWall,
            _settings$autoType = settings.autoType,
            autoType = _settings$autoType === undefined ? false : _settings$autoType,
            _settings$friction = settings.friction,
            friction = _settings$friction === undefined ? 0.2 : _settings$friction,
            _settings$restitution = settings.restitution,
            restitution = _settings$restitution === undefined ? 0 : _settings$restitution,
            _settings$density = settings.density,
            density = _settings$density === undefined ? 1.0 : _settings$density,
            _settings$maxHaptics = settings.maxHaptics,
            maxHaptics = _settings$maxHaptics === undefined ? -1 : _settings$maxHaptics,
            _settings$follow = settings.follow,
            follow = _settings$follow === undefined ? null : _settings$follow,
            _settings$deflation = settings.deflation,
            deflation = _settings$deflation === undefined ? 0.5 : _settings$deflation;


        var initialPos = position.clone();

        var edgeInfo = {};
        var outerEdges = [];

        var pixiGraphic = new PIXI.Graphics();
        var dropShadow = void 0; // a separate entity that follows this one with a slight offset
        var outerPolygon = void 0; // used for drawing the border

        var filter = isWall ? wallFilter : terrainFilter;

        // If the shape was not sliced we can draw using the tiled edges directly
        var useTiledEdges = function useTiledEdges() {
            outerPolygon = tiled.polygon.map(function (v) {
                return new Vector2(v.x, v.y);
            });
            repaint();
        };

        // But if the shape was sliced, we need to identify the edges ourselves
        var identifyEdges = function identifyEdges() {
            edgeInfo = {};
            outerEdges = [];
            body.fixtures.forEach(function (f) {
                var vertices = f.shape.GetVertices().map(Box2D.toBento);
                Utils.forEachEdge(vertices, function (v1, v2) {
                    var key = Utils.getEdgeKey(v1, v2);
                    var edge = edgeInfo[key];
                    if (!edge) {
                        edge = { v1: v1, v2: v2, count: 0 };
                        edgeInfo[key] = edge;
                    }
                    edge.count++;
                });
            });
            for (var key in edgeInfo) {
                var edge = edgeInfo[key];
                if (edge.count === 1) {
                    outerEdges.push(edge);
                }
            }
            buildOuterPolygon();
            repaint();
        };
        var isOuterEdge = function isOuterEdge(v1, v2) {
            var key = v1.x + ',' + v1.y + ':' + v2.x + ',' + v2.y;
            var edge = edgeInfo[key];
            return edge && edge.count === 1;
        };

        /**
         * Helper function for slicing.
         * Returns true if the given point lies on an outer edge.
         * (inner edges are those that are shared by two shapes in this body)
         */
        var onOuterEdge = function onOuterEdge(p) {
            for (var i = 0; i < outerEdges.length; i++) {
                if (Utils.pointOnEdge(p, outerEdges[i].v1, outerEdges[i].v2)) {
                    return true;
                }
            }
            return false;
        };

        var buildOuterPolygon = function buildOuterPolygon() {
            var edges = outerEdges.slice();
            for (var i = 0; i < edges.length - 1; i++) {
                var edge = edges[i];
                for (var j = i + 1; j < edges.length; j++) {
                    var nextEdge = edges[j];
                    if (Utils.approxEqualVec(edge.v2, nextEdge.v1, 0.001)) {
                        edges[j] = edges[i + 1];
                        edges[i + 1] = nextEdge;
                        break;
                    } else if (Utils.approxEqualVec(edge.v2, nextEdge.v2, 0.001)) {
                        var _ref = [nextEdge.v2, nextEdge.v1];
                        nextEdge.v1 = _ref[0];
                        nextEdge.v2 = _ref[1];

                        edges[j] = edges[i + 1];
                        edges[i + 1] = nextEdge;
                        break;
                    }
                }
                // this error isn't true? In fact breaking here causes incorrect behavior.
                // Utils.log(`Couldn't find next node. Body has two disconnected shapes??`);
                // break;
            }
            outerPolygon = [];
            edges.forEach(function (edge) {
                outerPolygon.push(edge.v1);
            });
        };

        var repaint = function repaint() {
            var pixiPolygon = void 0;
            if (isWall || deflation === 0) {
                pixiPolygon = new PIXI.Polygon(outerPolygon.map(function (v) {
                    return new PIXI.Point(v.x, v.y);
                }));
            } else {
                var points = [];
                Utils.forEachCorner(outerPolygon, function (p2, p1, p3) {
                    var diff1 = p2.subtract(p1).normalize();
                    var diff2 = p3.subtract(p1).normalize();
                    var deflate = diff1.addTo(diff2).normalize().scale(deflation);
                    points.push(p1.x + deflate.x);
                    points.push(p1.y + deflate.y);
                });
                pixiPolygon = new PIXI.Polygon(points);
            }

            pixiGraphic = new PIXI.Graphics();

            var textureName = void 0;
            var scale = void 0,
                angle = void 0;
            var shadowColor = void 0,
                shadowAlpha = void 0;
            var lineColor = void 0;
            if (isWall) {
                if (type === 'dynamic') {
                    scale = 0.08;
                    textureName = 'texture/solid-wave-cyan';
                    lineColor = 0x4a90de;
                } else {
                    var _getThemeInfo = getThemeInfo();

                    scale = _getThemeInfo.scale;
                    var _getThemeInfo$angle = _getThemeInfo.angle;
                    angle = _getThemeInfo$angle === undefined ? 0 : _getThemeInfo$angle;
                    lineColor = _getThemeInfo.lineColor;
                    textureName = _getThemeInfo.textureName;
                }
                shadowColor = 0x0072ff;
                shadowAlpha = 0.43;
            } else {
                scale = 0.5;
                textureName = 'texture/slice';
                shadowColor = 0x0072ff;
                shadowAlpha = 0.21;
            }

            var texture = Utils.newTexture(textureName);

            // create mesh for inner fill
            var vertices = new Float32Array(pixiPolygon.points);
            var indices = new Uint16Array(earcut(vertices));
            var uvs = new Float32Array(vertices);
            for (var i = 0; i < vertices.length; i += 2) {
                var vec = new Vector2(uvs[i], uvs[i + 1]);
                if (angle) {
                    vec.rotateRadian(angle);
                }
                uvs[i] = (vec.x + initialPos.x) / (texture.width * scale);
                uvs[i + 1] = (vec.y + initialPos.y) / (texture.height * scale);
                // uvs[i] = (uvs[i] + initialPos.x) / (texture.width * scale);
                // uvs[i+1] = (uvs[i+1] + initialPos.y) / (texture.height * scale);
            }
            var mesh = new PIXI.mesh.Mesh(texture, vertices, uvs, indices, PIXI.mesh.Mesh.DRAW_MODES.TRIANGLES);
            pixiGraphic.addChild(mesh);

            // create drop shadow
            var shadow = new PIXI.Graphics();
            shadow.beginFill(shadowColor, shadowAlpha);
            shadow.drawPolygon(pixiPolygon);
            shadow.endFill();
            dropShadow.setGraphics(shadow);

            // outer border
            if (isWall) {
                var border = new PIXI.Graphics();
                border.lineStyle(1, lineColor);
                border.moveTo(outerPolygon[0].x, outerPolygon[0].y);
                Utils.forEachEdge(outerPolygon, function (v1, v2) {
                    var g1 = initialPos.add(v1);
                    var g2 = initialPos.add(v2);
                    var key = Utils.getEdgeKey(g1, g2);
                    var count = globalEdges[key] || 0;
                    if (count > 1) {
                        border.moveTo(v2.x, v2.y);
                    } else {
                        border.lineTo(v2.x, v2.y);
                    }
                });
                pixiGraphic.addChild(border);
            }
        };

        var paintEllipse = function paintEllipse() {
            outerPolygon = [];
            var points = Utils.clamp(8, Math.floor(tiled.width) * 2, 32);
            var angleShift = Math.PI * 2 / points;
            var phi = 0.0;
            var coords = [];
            var radius = tiled.width / 2;
            for (var i = 0; i < points; i++) {
                outerPolygon.push(new Vector2(radius + radius * Math.cos(phi), radius + radius * Math.sin(phi)));
                phi += angleShift;
            }
            repaint();
        };

        // Become static/kinematic, if we're overlapping some wall which is also static/kinematic.
        // This is useful to enforce the rule that split terrain should remain anchored if it's still touching something solid.
        // But this alone is not enough to have e.g. sliceable terrain attached to a trolley.  (probably need to use Weld joints for that)
        var chooseType = function chooseType() {
            var transform = body.getB2Body().GetTransform();
            var hitType = null;
            for (var i = 0; i < body.fixtures.length; i++) {
                var shape = body.fixtures[i].shape;
                Box2D.queryShape(shape, transform, function (hitFixture) {
                    var t = hitFixture.body.getType();
                    if (hitFixture.filter === wallFilter && (t === 'static' || t === 'kinematic')) {
                        hitType = t;
                        return false;
                    }
                    return true;
                });
                if (hitType) {
                    break;
                }
            }
            body.setType(hitType || 'dynamic');
            setupHaptics();
        };

        var didSetupHaptics = false;
        var setupHaptics = function setupHaptics() {
            if (didSetupHaptics) return;
            didSetupHaptics = true;
            body.fixtures.forEach(function (f) {
                if (!f.onPostSolve) {
                    f.onPreSolve = hapticPreSolve;
                    f.onPostSolve = hapticPostSolve;
                }
            });
        };

        var linearVelSqrMagnitude = 0;
        var angularVel = 0;

        var prevLinearVelSqrMagnitude = 0;
        var prevAngularVel = 0;

        var hapticPreSolve = function hapticPreSolve(other, contact, oldManifold) {
            linearVelSqrMagnitude = body.getLinearVelocity().sqrMagnitude();
            angularVel = body.getAngularVelocity();
        };

        var hapticPostSolve = function hapticPostSolve(other, contact, impulse) {
            if (maxHaptics === 0) {
                return;
            }
            var velDiff = Math.abs(linearVelSqrMagnitude - prevLinearVelSqrMagnitude);
            var angDiff = Math.abs(angularVel - prevAngularVel);
            prevLinearVelSqrMagnitude = linearVelSqrMagnitude;
            prevAngularVel = angularVel;

            // console.log('vel', velDiff.toFixed(3), 'ang', angDiff.toFixed(3), 'mag', Math.sqrt(linearVelSqrMagnitude).toFixed(3));
            if (other.entity.name !== 'terrain' || Bento.objects.timer < nextAllowedHapticTime || !(linearVelSqrMagnitude > 20 * 20 || Math.abs(angularVel) > 0.1)
            // || !(velDiff > 10 || angDiff > 0.05)
            ) {
                    return;
                }

            var amount = impulse.normalImpulses[0];
            // console.log(amount);
            if (amount > 0.1) {
                Bento.audio.playSound('sfx_object_hits');
            }
            if (amount > 5) {
                TapticEngine.impact({ type: 'medium' });
            } else if (amount > 0.1) {
                TapticEngine.impact({ type: 'light' });
            } else {
                return;
            }
            // console.log('haptic', amount, linearVelSqrMagnitude, Math.abs(angularVel));
            nextAllowedHapticTime = Math.max(nextAllowedHapticTime, Bento.objects.timer + 4);
            if (maxHaptics > 0) {
                // console.log('DEC', entity.id);
                maxHaptics--;
            }
        };

        // This leads to very suboptimal terrain (consisting of many triangle polygon shapes)
        // But reduces the chance of glitchiness such as clipping into walls
        // Such glitchiness only seems to occur if a resulting shape contains collinear points, so it's better to just triangulate those shapes.
        var triangulateAll = false;

        if (tiled) {

            position = new Vector2(tiled.x, tiled.y);
            rotation = tiled.rotation;

            if (tiled.polygon) {
                var concavePoly = tiled.polygon.map(function (v) {
                    return [v.x, v.y];
                });
                decomp.makeCCW(concavePoly);
                var convexPolys = void 0;

                if (triangulateAll) {
                    var triangles = earcut(Utils.flatten(concavePoly));
                    convexPolys = [];
                    for (var i = 0; i < triangles.length; i += 3) {
                        convexPolys.push([concavePoly[triangles[i]], concavePoly[triangles[i + 1]], concavePoly[triangles[i + 2]]]);
                    }
                } else {
                    convexPolys = decomp.quickDecomp(concavePoly);
                    Utils.triangulateBadPolys(convexPolys);
                }

                fixtures = convexPolys.map(function (poly) {
                    return {
                        filter: filter,
                        shape: new Box2D.PolygonShape(poly.map(function (v) {
                            return new Vector2(v[0], v[1]);
                        })),
                        friction: friction,
                        restitution: restitution,
                        density: density
                    };
                });
            } else if (tiled.ellipse) {
                if (!isWall) {
                    Utils.log('Circle shapes are not supposed to be sliceable.');
                }
                if (tiled.width !== tiled.height) {
                    Utils.log('Circle width and height should match.');
                }
                fixture = {
                    filter: filter,
                    shape: new Box2D.CircleShape(tiled.width / 2, new Vector2(tiled.width / 2, tiled.height / 2)),
                    friction: friction,
                    restitution: restitution,
                    density: density
                };
            } else {
                // assume box shape
                var vertices = [new Vector2(tiled.width, 0), new Vector2(tiled.width, tiled.height), new Vector2(0, tiled.height), new Vector2(0, 0)];
                fixture = {
                    filter: filter,
                    shape: new Box2D.PolygonShape(vertices),
                    friction: friction,
                    restitution: restitution,
                    density: density
                };
            }
        }

        // may have been passed a single fixture or a list
        if (!fixtures) {
            fixtures = fixture ? [fixture] : [];
        }

        if (follow) {
            type = 'kinematic';
        }

        // let hapticPos = new Vector2(0, 0);
        // let hapticOldPos = new Vector2(0, 0);
        // let hapticRot = 0;
        // let hapticOldRot = 0;

        /*
        let makeHapticCollisionCallback = function () {
            return (other, contact, impulse) => {
                if (Bento.objects.timer < nextAllowedHapticTime) {
                    return;
                }
                if (other.entity.name === 'terrain') {
                    let mag = body.getLinearVelocity().magnitude();
                    let mass = body.getMass();
                    let momentum = mag*mass;
                    var amount = impulse.normalImpulses[0];
                    // var amount = oldManifold.m_points[0].m_points[0].m_normalImpulse;
                    console.log(amount);
                    if (amount > 5) {
                        if (momentum > 14) {
                            TapticEngine.impact({ type: 'light' });
                            console.log('light', momentum);
                        } else if (momentum > 40) {
                            TapticEngine.impact({ type: 'medium' });
                            console.log('medium', momentum);
                        } else if (momentum > 250) {
                            TapticEngine.impact({ type: 'heavy' });
                            console.log('heavy', momentum);
                        } else {
                            console.log('none', mag, mass,momentum);
                            return;
                        }
                        nextAllowedHapticTime = Math.max(nextAllowedHapticTime, Bento.objects.timer+4);
                    }
                }
            };
        };
        */

        var body = new Box2D.Body({
            type: type,
            fixtures: fixtures
            // bullet: true,
            // allowSleep: false,
        });

        var behavior = {
            name: 'behavior',
            start: function start() {
                dropShadow = new DropShadow({
                    terrain: entity
                });
                Bento.objects.attach(dropShadow);
            },
            destroy: function destroy() {
                Bento.objects.remove(dropShadow);
            },
            draw: function draw(_ref2) {
                var renderer = _ref2.renderer;

                // body.fixtures.forEach(f => Box2D.drawShape(renderer, f.shape, seamColor));
                // outerEdges.forEach(({v1,v2})=> {
                //     renderer.drawLine(color, v1.x, v1.y, v2.x, v2.y, 1);
                // });
                renderer.drawPixi(pixiGraphic);
                // Utils.forEachEdge(outerPolygon, (v1, v2) => {
                // renderer.drawLine(color, v1.x, v1.y, v2.x, v2.y, 1);
                // });
            }
        };

        var postBodyInit = {
            name: 'postBodyInit',
            start: function start() {
                if (tiled) {
                    if (tiled.polygon) {
                        useTiledEdges();
                    } else if (tiled.ellipse) {
                        paintEllipse();
                    } else {
                        identifyEdges(); // rectangle shape
                    }
                } else {
                    identifyEdges(); // sliced shape
                }
                if (autoType) {
                    chooseType();
                }
                if (body.getType() === 'dynamic') {
                    setupHaptics();
                }
            },
            update: function update() {
                dropShadow.position.x = entity.position.x + 1.5;
                dropShadow.position.y = entity.position.y + 1.5;
                dropShadow.rotation = entity.rotation;
            }
        };

        var getInheritableSettings = function getInheritableSettings() {
            return {
                maxHaptics: maxHaptics, // How many times can this entity produce haptic feedback and 'bump' SFX?.
                deflation: deflation // How much to shrink the rendered polygon
            };
        };

        var getType = function getType() {
            return body.getType();
        };

        var entity = new Entity({
            z: isWall ? Globals.Layers.WALL : Globals.Layers.TERRAIN,
            name: name || tiled && tiled.name || 'terrain',
            family: ['terrain'],
            position: position,
            rotation: rotation,
            components: [behavior, body, postBodyInit]
        }).extend({
            isWall: isWall,
            isOuterEdge: isOuterEdge,
            identifyEdges: identifyEdges,
            chooseType: chooseType,
            getType: getType,
            getInheritableSettings: getInheritableSettings
        });

        if (follow) {
            entity.attach(new PathFollower({
                pathName: follow
            }));
        }

        return entity;
    };

    Utils.extend(Terrain, {
        filter: terrainFilter,
        clearGlobalEdges: clearGlobalEdges,
        addGlobalEdge: addGlobalEdge
    });

    var _export_default = Terrain;


    var DropShadow = function DropShadow(settings) {
        var terrain = settings.terrain;

        var graphics = null;
        var behavior = {
            name: 'behavior',
            update: function update() {
                // entity.position.x = terrain.position.x + 1.5;
                // entity.position.y = terrain.position.y + 1.5;
                // entity.rotation = terrain.rotation;
            },
            draw: function draw(_ref3) {
                var renderer = _ref3.renderer;

                if (graphics) {
                    renderer.drawPixi(graphics);
                }
            }
        };
        var entity = new Entity({
            z: Globals.Layers.TERRAIN_SHADOW,
            name: 'dropShadow',
            position: terrain.position.clone(),
            components: [behavior]
        }).extend({
            setGraphics: function setGraphics(gfx) {
                graphics = gfx;
            }
        });
        return entity;
    };
    return _export_default;
});