'use strict';

bento.define('gui/chestprogress', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/entity', 'bento/eventsystem', 'bento/gui/text', 'bento/utils', 'bento/screen', 'bento/tween', 'globals', 'bento/components/eventlistener', 'modules/savemanager', 'modules/levelmanager', 'entities/lemming'], function (Bento, Vector2, Rectangle, Sprite, Entity, EventSystem, Text, Utils, Screen, Tween, Globals, EventListener, SaveManager, LevelManager, Lemming) {

    /**
     * If true, progress is stored in save file.
     * If false, progress is directly tied to which level the player is currently on.
     */
    var freeProgression = true;

    /**
     * Get which 'segment' of the progression bar we are on
     * Returns a number from 0..4, where:
     * -1: player just started the game or claimed a chest
     * 0: first segment
     * 4: final segment (the one with the chest on it)
     * 5+: player didn't watch the ad so they are still at the chest! (in freeProgression mode only)
     */
    var getProgress = function getProgress() {
        // return 4;
        return freeProgression ? SaveManager.load('chestProgress') : Utils.currentLevel % 5;
    };

    /**
     * True if the player is on the last segment of the bar.
     * Be sure the call incrementProgress before checking this!
     */
    var hasChestAvailable = function hasChestAvailable() {
        return getProgress() >= 4;
    };

    /**
     * Call this after the player beat a level
     */
    var incrementProgress = function incrementProgress() {
        if (!freeProgression) {
            return;
        }
        var progress = SaveManager.load('chestProgress');
        progress++;
        // if (progress > 4) progress = 4;
        SaveManager.save('chestProgress', progress);
    };

    /**
     * Call this after the player watched the ad to spin the wheel.
     */
    var resetProgress = function resetProgress() {
        if (!freeProgression) {
            return;
        }
        SaveManager.save('chestProgress', -1);
    };

    var ChestProgress = function ChestProgress(settings) {
        var position = settings.position,
            scale = settings.scale,
            onProgressed = settings.onProgressed;


        // let progress = (LevelManager.getNumberCleared()-1) % 5;
        var progress = getProgress();
        var currentSegment = Utils.clamp(0, progress, 4);

        // segments

        var segments = [];

        var makeSegment = function makeSegment(n) {
            var isFilled = n < progress;
            var sprite = new Sprite({
                imageName: 'gui/progression-' + (isFilled ? 'full' : 'empty'),
                originRelative: new Vector2(0.5, 0.5)
            });
            var segment = new Entity({
                name: 'progressSegment',
                position: new Vector2((n - 2) * 32, 11),
                scale: new Vector2(0.16, 0.16),
                components: [sprite]
            });
            segments.push(segment);
        };
        for (var i = 0; i < 5; i++) {
            makeSegment(i);
        }

        // chest

        var chestSprite = new Sprite({
            imageName: 'gui/chest-icon',
            originRelative: new Vector2(0.5, 0.5),
            frameCountX: 1,
            frameCountY: 1,
            animations: {
                default: {
                    speed: 0,
                    frames: [0]
                }
            }
        });
        var chestIcon = new Entity({
            name: 'chestIcon',
            position: new Vector2(64, -5),
            // scale: new Vector2(0.05, 0.05),
            scale: new Vector2(0.15, 0.15),
            components: [chestSprite]
        });

        // player

        var currentSkin = SaveManager.load('character');
        var anims = Lemming.getAnimations(currentSkin);
        var currentAnim = 'idle';
        var sprite = new Sprite(anims['idle']);

        var lemming = new Entity({
            name: 'progressLemming',
            position: new Vector2(segments[currentSegment].position.x, -2),
            scale: new Vector2(0.8, 0.8),
            components: [sprite]
        });

        var setLemmingAnim = function setLemmingAnim(animName) {
            currentAnim = animName;
            sprite.setup(anims[animName]);
        };

        var onCharacterChanged = function onCharacterChanged(characterId) {
            if (currentSkin !== characterId) {
                currentSkin = characterId;
                anims = Lemming.getAnimations(currentSkin);
                sprite.setup(anims[currentAnim]);
            }
        };

        var behavior = {
            name: 'behavior',
            start: function start() {
                // player may have changed skin since the last time we attached this instance
                onCharacterChanged(SaveManager.load('character'));
                EventSystem.on('characterChanged', onCharacterChanged);
            },
            destroy: function destroy() {
                EventSystem.off('characterChanged', onCharacterChanged);
            }
        };

        var entity = new Entity({
            name: 'chestProgress',
            position: position,
            scale: scale,
            components: [behavior, chestIcon, lemming].concat(segments)
        });

        var bounceSegment = function bounceSegment() {
            var bouncePlayer = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
            var bounceChest = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

            var seg = segments[currentSegment];
            var segY = seg.position.y;
            var bounceTime = 16;
            seg.attach(Tween.TweenBehavior({
                from: seg.scale.x * 1.1,
                to: seg.scale.x,
                in: bounceTime,
                ease: 'easeOutQuad',
                onStart: function onStart() {
                    seg.getComponent('sprite', function (s) {
                        s.setup({
                            imageName: 'gui/progression-full',
                            originRelative: new Vector2(0.5, 0.5)
                        });
                    });
                    if (onProgressed) {
                        onProgressed();
                    }
                },
                onUpdate: function onUpdate(v, t) {
                    t = Utils.clamp(0, t, bounceTime);
                    seg.scale.x = v;
                    seg.scale.y = v;
                    seg.position.y = segY + Math.sin(t / bounceTime * Math.PI) * 4;
                    if (bouncePlayer) {
                        lemming.transform.y = Math.sin(t / bounceTime * Math.PI) * 4;
                    }
                    if (bounceChest) {
                        chestIcon.transform.y = Math.sin(t / bounceTime * Math.PI) * 4;
                    }
                },
                onComplete: function onComplete() {}
            }));
        };

        var lemmingDoVictoryAnim = function lemmingDoVictoryAnim() {
            setLemmingAnim('victory');
            var frame = 0.0,
                oldFrame = 0.0;
            var lemmingPos = lemming.position.clone();
            var chestPos = chestIcon.position.clone();
            lemming.attach({
                name: 'jumpBehavior',
                update: function update() {
                    oldFrame = frame;
                    frame = sprite.getCurrentFrame();
                    if (frame >= 2.0 && oldFrame < 2.0) {
                        // if we passed frame 2, jump!
                        lemming.attach(Tween.TweenBehavior({
                            from: 0, to: Math.PI, in: 16,
                            ease: 'linear',
                            onUpdate: function onUpdate(v, t) {
                                lemming.position.y = lemmingPos.y - Math.sin(v) * 4;
                            },
                            onComplete: function onComplete() {
                                // upon landing, make the chest and player sink down a little
                                lemming.attach(Tween.TweenBehavior({
                                    from: 0, to: Math.PI, in: 10,
                                    ease: 'linear',
                                    onUpdate: function onUpdate(v, t) {
                                        lemming.position.y = lemmingPos.y + Math.sin(v) * 2;
                                        chestIcon.position.y = chestPos.y + Math.sin(v) * 1;
                                    }
                                }));
                            }
                        }));
                    }
                }
            });
        };

        // move the player
        switch (progress) {
            case 0:

                // falling onto the first segment doesn't look so good without rotation (and masking breaks if we use it)

                // entity.attach(Tween.TweenBehavior({
                //     delay: 40,
                //     from: lemming.position.add(Vector2(0, -24)),
                //     to: lemming.position.clone(),
                //     in: 18,
                //     ease: 'easeInQuad',
                //     applyOnDelay: true,
                //     onStart () {
                //         sprite.setup(anims['fall']);
                //     },
                //     onUpdate (v, t) {
                //         lemming.position.x = v.x;
                //         lemming.position.y = v.y;
                //     },
                //     onComplete () {
                //         sprite.setup(anims['idle']);
                //         bounceSegment(true);
                //     }
                // }));
                // break;

                // so we can make the player do a barrel roll instead ;-)

                var firstWalkTime = 24;
                entity.attach(Tween.TweenBehavior({
                    delay: 40,
                    from: lemming.position.add(Vector2(-40, 0)),
                    to: lemming.position.clone(),
                    in: firstWalkTime,
                    ease: 'linear',
                    applyOnDelay: true,
                    onStart: function onStart() {
                        setLemmingAnim('walk');
                    },
                    onUpdate: function onUpdate(v, t) {
                        t = Utils.clamp(0, t, firstWalkTime);
                        lemming.position.x = v.x;
                        lemming.position.y = v.y - Math.sin(t / firstWalkTime * Math.PI) * 5;
                        lemming.rotation = t / firstWalkTime * Math.PI * 2;
                    },
                    onComplete: function onComplete() {
                        setLemmingAnim('idle');
                        bounceSegment(true);
                    }
                }));
                break;

            case 1:
            case 2:
            case 3:
                // player walks to next segment
                var walkTime = 22;
                entity.attach(Tween.TweenBehavior({
                    delay: 40,
                    from: lemming.position.add(Vector2(-32, 0)),
                    to: lemming.position.clone(),
                    in: walkTime,
                    ease: 'linear',
                    applyOnDelay: true,
                    onStart: function onStart() {
                        setLemmingAnim('walk');
                    },
                    onUpdate: function onUpdate(v, t) {
                        lemming.position.x = v.x;
                        lemming.position.y = v.y - Math.sin(t / walkTime * Math.PI) * 5;
                    },
                    onComplete: function onComplete() {
                        setLemmingAnim('idle');
                        bounceSegment(true);
                    }
                }));
                break;

            case 4:
                // player jumps to chest

                var jumpTime = 22;
                entity.attach(Tween.TweenBehavior({
                    delay: 40,
                    from: lemming.position.add(Vector2(-32, 0)),
                    to: lemming.position.add(Vector2(0, -14.5)),
                    in: jumpTime,
                    ease: 'linear',
                    applyOnDelay: true,
                    onStart: function onStart() {
                        // remake the sprite so masking no longer applies to the player
                        lemming.remove(sprite);
                        sprite = new Sprite(anims['walk']);
                        currentAnim = 'walk';
                        lemming.attach(sprite);
                    },
                    onUpdate: function onUpdate(v, t) {
                        lemming.position.x = v.x;
                        lemming.position.y = v.y - Math.sin(t / jumpTime * Math.PI) * 8;
                    },
                    onComplete: function onComplete() {
                        bounceSegment(true, true);
                        lemmingDoVictoryAnim();
                    }
                }));
                break;

            default:
                // 5+, player remained at the chest because they didn't open it

                var jumpTime2 = 22;
                var prevT = 0;
                entity.moveComponentTo(lemming, 0);
                entity.attach(Tween.TweenBehavior({
                    delay: 40,
                    from: lemming.position.add(Vector2(0, 0)),
                    to: lemming.position.add(Vector2(0, -14.5)),
                    in: jumpTime2,
                    ease: 'linear',
                    applyOnDelay: true,
                    onStart: function onStart() {
                        // remake the sprite so masking no longer applies to the player
                        lemming.remove(sprite);
                        sprite = new Sprite(anims['walk']);
                        currentAnim = 'walk';
                        lemming.attach(sprite);
                    },
                    onUpdate: function onUpdate(v, t) {
                        lemming.position.x = v.x;
                        lemming.position.y = v.y - Math.sin(t / jumpTime2 * Math.PI) * 8;
                        if (t >= 11 && prevT < 11) {
                            entity.moveComponentTo(lemming, 2);
                        }
                        prevT = t;
                    },
                    onComplete: function onComplete() {
                        bounceSegment(true, true);
                        lemmingDoVictoryAnim();
                    }
                }));
        }

        return entity;
    };

    Utils.extend(ChestProgress, {
        hasChestAvailable: hasChestAvailable,
        getProgress: getProgress,
        incrementProgress: incrementProgress,
        resetProgress: resetProgress
    });

    return ChestProgress;
});