'use strict';

bento.define('gui/gui', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/components/clickable', 'bento/entity', 'bento/eventsystem', 'entities/gui/gamecounter', 'entities/gui/gamebutton', 'entities/gui/gametext', 'bento/gui/text', 'bento/utils', 'bento/screen', 'bento/tween', 'globals', 'bento/components/fill', 'bento/components/eventlistener', 'components/wiggle', 'components/hdsprite', 'gui/uiloader', 'modules/textstyles', 'modules/confetti', 'modules/tapticengine', 'entities/gui/starsearned'], function (Bento, Vector2, Rectangle, Sprite, Clickable, Entity, EventSystem, GameCounter, GameButton, GameText, Text, Utils, Screen, Tween, Globals, Fill, EventListener, Wiggle, HDSprite, UILoader, TextStyles, Confetti, TapticEngine, StarsEarned) {

    var scaleFactor = void 0,
        scaleFactorTall = void 0,
        notch = void 0;

    var updateScaleFactors = function updateScaleFactors() {

        var viewport = Bento.getViewport();
        var realViewport = Bento.getViewport(true);

        // has a value of 1.0 on iPhone 6/7/8, gets smaller as the device gets fatter
        scaleFactor = Math.min(1.0, viewport.height / viewport.width / (320 / 180));

        // has a value of 1.0 on iPhone X, gets smaller as the device gets fatter
        scaleFactorTall = Math.min(1.0, viewport.height / viewport.width / (390 / 180));

        // true for iPhone X, regardless of whether a banner ad is shown.
        notch = realViewport.height / realViewport.width / (390 / 180) >= 1.0;

        Gui.scaleFactor = scaleFactor;
        Gui.scaleFactorTall = scaleFactorTall;
        Gui.notch = notch;
    };

    Bento.objects.attach(new EventListener({
        eventName: 'bannerVisibilityChanged',
        global: true,
        ignorePause: true,
        onEvent: function onEvent(data) {
            updateScaleFactors();
        }
    }));

    /**
     * @snippet makeButton|snippet
    makeButton(obj, {
        shadow: true,
        onClick () {
            
        }
    });
     */
    var makeButton = function makeButton(obj, settings) {
        var _settings$canClick = settings.canClick,
            canClick = _settings$canClick === undefined ? function () {
            return true;
        } : _settings$canClick,
            onClick = settings.onClick,
            onButtonDown = settings.onButtonDown,
            onButtonUp = settings.onButtonUp,
            _settings$shadow = settings.shadow,
            shadow = _settings$shadow === undefined ? false : _settings$shadow,
            _settings$haptics = settings.haptics,
            haptics = _settings$haptics === undefined ? true : _settings$haptics,
            _settings$depress = settings.depress,
            depress = _settings$depress === undefined ? true : _settings$depress,
            _settings$sound = settings.sound,
            sound = _settings$sound === undefined ? true : _settings$sound;


        if (shadow) {
            var shadowSprite = new HDSprite({
                name: 'shadow',
                imageName: Utils.isString(shadow) ? shadow : 'gui/button-shadow',
                gamepixelDimension: obj.dimension.clone(),
                originRelative: new Vector2(0.5, 0.5)
            });
            shadowSprite.position.x = 1.6;
            shadowSprite.position.y = 1.6;
            obj.attach(shadowSprite);
            obj.moveComponentTo(shadowSprite, 1);
        }

        var isClicking = false;
        obj.attach(new Clickable({
            onClick: function onClick(data) {
                // if (!obj.active) return;
                if (!canClick(data)) return;
                if (depress) obj.transform.y = 1;
                isClicking = true;
                if (onButtonDown) onButtonDown(data);
                if (haptics) TapticEngine.impact({ type: 'light' });
                if (sound) Bento.audio.playSound('sfx_button');
            },
            onHoldLeave: function onHoldLeave() {
                obj.transform.y = 0;
                isClicking = false;
            },
            pointerUp: function pointerUp() {
                obj.transform.y = 0;
            },
            onClickUp: function onClickUp(data) {
                if (onButtonDown) onButtonUp(data);
                if (isClicking) {
                    if (onClick && canClick(data)) {
                        onClick(data);
                        EventSystem.fire('GameAnalytics-addDesignEvent', {
                            eventId: Utils.currentScreenName() + ':' + obj.name + ':pressed'
                        });
                    }
                    isClicking = false;
                }
            }
        }));
    };

    /**
     * @snippet makeText|snippet
    makeText(obj, {
        text: Localization.getText('${1:text}'),
        fontSettings: TextStyles.get('${2:main}'),
        align: 'center',
    })
     */
    var makeText = function makeText(obj, settings) {
        var name = obj.name,
            position = obj.position,
            dimension = obj.dimension;

        var upscale = 2;
        var text = new Text(Utils.extend({
            maxWidth: dimension.width,
            maxHeight: dimension.height,
            sharpness: 4 * upscale
        }, settings));

        // Assume the first sub-entity is a 'textplacement' object which we can replace with text
        var textplacement = obj.components.find(function (comp) {
            return comp.isEntity;
        });
        // text.scale = textplacement.scale.scale(1/upscale);
        text.transform.y += 1.0;
        if (settings.align === 'left') {
            text.transform.x -= upscale * dimension.width / 4; // wtf
        }
        obj.remove(textplacement);
        obj.attach(text);
    };

    var parents = {},
        objects = {};

    var Gui = {

        isRestoring: false, // true when Gui.restore is being used to re-attach a previously-active UI

        scaleFactor: scaleFactor,
        scaleFactorTall: scaleFactorTall,
        notch: notch,

        load: function load(settings) {
            var parent = settings.parent,
                handlers = settings.handlers,
                assetName = settings.assetName,
                onComplete = settings.onComplete;


            if (!parent) {
                // we are replacing the main UI
                objects = {};
                parents = {};
            }

            var parentFns = [];
            var handlerFns = [];

            if (handlers.init) {
                handlers.init(settings);
            }

            var uiLoader = new UILoader({
                loadJson: assetName,
                onObject: function onObject(obj) {

                    var isRoot = false;
                    if (obj.name === 'root') {
                        isRoot = true;
                        if (parent) {
                            // for a sub-gui, the original root is discarded, parent becomes the new root
                            obj.removeSelf();
                            obj.components.forEach(function (child) {
                                if (child.isEntity) {
                                    obj.remove(child);
                                    parent.attach(child);
                                }
                            });
                            obj = parent;
                        }
                    }

                    // schdule handlers to be called and parents to be assigned
                    objects[obj.name] = obj;
                    parentFns.push(function () {
                        return parents[obj.name] = obj.parent;
                    });
                    var handler = handlers[obj.name];
                    if (handler) {
                        handlerFns.push(function () {
                            return handler(obj);
                        });
                    }

                    // call handlers once root is fully added
                    if (isRoot) {
                        obj.attach({
                            name: 'rootHook',
                            start: function start() {
                                parentFns.forEach(function (f) {
                                    return f();
                                });
                                handlerFns.forEach(function (f) {
                                    return f();
                                });
                                if (!parent) {
                                    EventSystem.fire('guiLoaded');
                                }
                                if (onComplete) {
                                    onComplete();
                                }
                                obj.remove(this);
                            }
                        });
                    }
                }
            });
            Bento.objects.attach(uiLoader);
        },
        get: function get(name, fn) {
            var obj = objects[name];
            if (obj) {
                if (fn) fn(obj);
                return obj;
            }
            return null;
        },
        show: function show(name) {
            var obj = Utils.isString(name) ? objects[name] : name;
            if (!obj) {
                console.warn('Couldn\'t find UI object \'' + name + '\'');
                return;
            }
            var parent = parents[obj.name];
            if (!parent) {
                console.warn('UI object \'' + name + '\' doesn\'t have a designated parent.');
                return;
            }
            var showable = obj.getComponent('showable');
            if (showable) {
                if (!showable.isShown) {
                    showable.show(parent); // animate in
                }
            } else {
                parent.attach(obj);
            }
        },
        hide: function hide(name) {
            var obj = Utils.isString(name) ? objects[name] : name;
            if (!obj) {
                console.warn('Couldn\'t find UI object \'' + name + '\'');
                return;
            }
            var parent = parents[obj.name];
            if (!parent) {
                // Utils.log(`UI object '${name}' doesn't have a designated parent.`);
                obj.removeSelf();
                return;
            }
            var showable = obj.getComponent('showable');
            if (showable) {
                if (showable.isShown) {
                    showable.hide(parent); // animate out
                }
            } else {
                parent.remove(obj);
            }
        },
        attach: function attach(obj) {
            if (objects[obj.name]) {
                Utils.log('Gui already contains an element called ' + obj.name);
            }
            objects['root'].attach(obj);
            objects[obj.name] = obj;
        },
        remove: function remove(obj) {
            objects['root'].remove(obj);
            objects[obj.name] = null;
        },
        restoreState: function restoreState(state) {
            Gui.isRestoring = true;

            var oldRoot = objects['root'];
            objects = state.objects;
            parents = state.parents;
            var root = objects['root'];

            // Put the root into a state that allows it to be re-added to the world without messing up object positions
            var removePositioners = function removePositioners(obj) {
                obj.removeByName('positioner');
                obj.removeByName('adjuster');
                obj.components.forEach(function (c) {
                    if (c && c.isEntity) removePositioners(c);
                });
            };
            removePositioners(root);

            Bento.objects.remove(oldRoot);
            Bento.objects.attach(root);

            state.tweens.forEach(function (tweenBehavior) {
                if (!tweenBehavior.isAdded) {
                    Bento.objects.attach(tweenBehavior);
                    // console.log(tweenBehavior);
                }
            });

            Gui.isRestoring = false;

            // let completeAnimations = function (obj) {
            //     obj.getComponent('showable', showable => {
            //         if (showable.isShown) {
            //             showable.show();
            //         } else {
            //             showable.hide();
            //         }
            //     });
            // };
            // completeAnimations(root);
        },
        getState: function getState() {
            var tweens = Bento.objects.getObjects().filter(function (o) {
                return o && o.name === 'tweenBehavior' && !o.global;
            });
            return {
                objects: objects,
                parents: parents,
                tweens: tweens
            };
        },
        getRootPosition: function getRootPosition(obj) {
            var p1 = obj.getWorldPosition();
            var p2 = objects['root'].getWorldPosition();
            return p1.subtract(p2);
        },


        makeButton: makeButton,
        makeText: makeText,

        Showable: function Showable(settings) {
            var _show = settings.show,
                _hide = settings.hide,
                _settings$isShown = settings.isShown,
                isShown = _settings$isShown === undefined ? true : _settings$isShown;

            var tween = void 0;
            return {
                name: 'showable',
                isShown: isShown,
                entity: null,
                attached: function attached(_ref) {
                    var entity = _ref.entity;

                    this.entity = entity;
                    if (!isShown) {
                        entity.removeSelf();
                    }
                },
                show: function show(parent) {
                    if (tween) tween.removeSelf();
                    tween = _show(parent);
                    if (tween === false) {
                        return; // show() can return false to indicate that this was never shown
                    }
                    this.isShown = true;
                },
                hide: function hide(parent) {
                    if (tween) tween.removeSelf();
                    tween = _hide(parent);
                    if (tween === false) {
                        return; // hide() can return false to indicate that this was never hidden
                    }
                    this.isShown = false;
                }
            };
        },
        FadeShowable: function FadeShowable(settings) {
            var _show2 = settings.show,
                _hide2 = settings.hide,
                isShown = settings.isShown,
                _settings$showTime = settings.showTime,
                showTime = _settings$showTime === undefined ? 20 : _settings$showTime,
                _settings$hideTime = settings.hideTime,
                hideTime = _settings$hideTime === undefined ? 8 : _settings$hideTime,
                _settings$showDelay = settings.showDelay,
                showDelay = _settings$showDelay === undefined ? 0 : _settings$showDelay,
                _settings$hideDelay = settings.hideDelay,
                hideDelay = _settings$hideDelay === undefined ? 0 : _settings$hideDelay;

            var showable = Gui.Showable({
                isShown: isShown,
                show: function show(parent) {
                    return new Tween({
                        from: 0, to: 1, in: showTime, delay: showDelay,
                        ease: 'easeOutQuad',
                        onStart: function onStart() {
                            if (_show2) _show2();
                            parent.attach(showable.entity);
                        },
                        onUpdate: function onUpdate(v, t) {
                            showable.entity.alpha = v;
                        }
                    });
                },
                hide: function hide(parent) {
                    return new Tween({
                        from: 1, to: 0, in: hideTime, delay: hideDelay,
                        ease: 'easeInQuad',
                        onStart: function onStart() {
                            if (_hide2) _hide2();
                        },
                        onUpdate: function onUpdate(v, t) {
                            showable.entity.alpha = v;
                        },
                        onComplete: function onComplete() {
                            parent.remove(showable.entity);
                        }
                    });
                }
            });
            return showable;
        },
        isShown: function isShown(name) {
            var obj = Utils.isString(name) ? objects[name] : name;
            if (!obj) {
                console.warn('Couldn\'t find UI object \'' + name + '\'');
                return false;
            }
            var showable = obj.getComponent('showable');
            return showable ? showable.isShown : obj.parent && obj.visible;
        },
        hasPosition: function hasPosition(mousePosition) {
            var eachEntity = function eachEntity(entity, f) {
                if (entity && entity.isEntity) {
                    var arr = entity.components;
                    for (var i = 0; i < arr.length; i++) {
                        eachEntity(arr[i], f);
                    }
                    f(entity);
                }
            };
            var hasPosition = false;
            eachEntity(Gui.get('root'), function (entity) {
                var clickable = entity.getComponent('clickable');
                var pos = entity.toComparablePosition(mousePosition);
                if (clickable && entity.getBoundingBox().hasPosition(pos)) {
                    hasPosition = true;
                }
            });
            return hasPosition;
        },
        adjust: function adjust(obj, settings) {
            var _settings$x = settings.x,
                x = _settings$x === undefined ? 0 : _settings$x,
                _settings$y = settings.y,
                y = _settings$y === undefined ? 0 : _settings$y,
                _settings$scale = settings.scale,
                scale = _settings$scale === undefined ? 1 : _settings$scale,
                onAdjust = settings.onAdjust;


            var hasAdjusted = false;

            obj.scale.x *= scale;
            obj.scale.y *= scale;
            obj.attach({
                name: 'adjuster',
                start: function start() {
                    // window.addEventListener('resize', this.adjust, false);
                    if (!hasAdjusted) {
                        this.adjust();
                    }
                },
                destroy: function destroy() {
                    // window.removeEventListener('resize', this.adjust);
                },
                adjust: function adjust() {
                    obj.components.forEach(function (c) {
                        if (c && c.name === 'positioner') obj.remove(c);
                    });
                    obj.position.x += x;
                    obj.position.y += y;
                    hasAdjusted = true;
                    if (onAdjust) {
                        onAdjust();
                    }
                }
            });
        }
    };

    updateScaleFactors();

    return Gui;
});