'use strict';

/**
 * UI Loader from JSON, note that it loads modules asynchronously
 * @moduleName UILoader
 * @snippet UILoader|constructor
UILoader({
    assetName: '',
    onObject: function (object) {},
    onComplete: function () {}
})
 */
bento.define('gui/uiloader', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/components/clickable', 'bento/entity', 'bento/eventsystem', 'bento/gui/clickbutton', 'bento/gui/counter', 'bento/gui/text', 'bento/utils', 'bento/tween', 'bento/components/nineslice', 'components/hdnineslice', 'components/iconbutton', 'components/textplacement'], function (Bento, Vector2, Rectangle, Sprite, Clickable, Entity, EventSystem, ClickButton, Counter, Text, Utils, Tween, NineSlice, HDNineSlice, IconButton, TextPlacement) {
    'use strict';
    // cached modules to decrease async times

    var cachedModules = {};
    var didPreload = false;
    /**
     * Main module
     */
    var UILoader = function UILoader(settings) {
        // var viewport = Bento.getViewport();
        var assetName = settings.loadJson || "no asset name supplied";
        var onObject = settings.onObject;
        var json = settings.hasOwnProperty('json') ? settings.json : {};
        var uiLoader = {
            name: 'uiLoader',
            ui: {},
            start: function start(data) {
                if (Object.keys(json).length) {
                    uiLoader.ui = json;
                } else {
                    //if no json supplied, load json from file
                    if (Bento.assets.hasAsset(assetName, 'json')) {
                        uiLoader.ui = Bento.assets.getJson(assetName);
                    } else {
                        Utils.log('ERROR: ui json "' + assetName + '" could not be found');
                    }
                }
                if (Object.keys(uiLoader.ui).length) {
                    uiLoader.load();
                }
            },
            load: function load() {
                // parse UI
                parse(uiLoader.ui, function (instance) {
                    // attach the parent object
                    Bento.objects.attach(instance);
                });
            }
        };
        var parse = function parse(object, onParseComplete) {

            var moduleName = object.module;
            // loop through components first to instantiate them

            // at this point we need a clone of the object and parse it --> could be memory heavy???
            // var objectSettings = Utils.cloneJson(object);

            var components = object.components;
            var componentInstances = [];

            var toLoad = 0;
            var loaded = 0;
            var didLoop = false;

            var onReadyToInstantiate = function onReadyToInstantiate() {
                // instantiate the module with moduleName 
                var settingsObject = {};

                if (object.offset) {
                    object.offset = toVector2(object.offset);
                }
                var offset = object.offset ? object.offset : new Vector2(0, 0);
                if (object.originRelative) {
                    object.originRelative = toVector2(object.originRelative);
                }
                var onGetConstructor = function onGetConstructor(Constructor) {
                    var instance = new Constructor(settingsObject);
                    instance.name = object.name; //force name to be what it is in the json
                    if (onParseComplete) {
                        onParseComplete(instance);
                    }

                    // cache constructor(?)
                    cachedModules[moduleName] = Constructor;

                    if (onObject) {
                        onObject(instance);
                    }
                };

                // the settings object is object itself, but we need to swap the components with the componentInstances array
                // to do that we make a shallow copy of the settings object
                // and then assign components as the componentInstances
                Utils.forEach(object, function (item, key) {
                    if (key !== 'module' && key !== 'position' && key !== 'rotationDegrees' && key !== 'components') {
                        settingsObject[key] = item;
                    }
                });

                if (components) {
                    settingsObject.components = componentInstances;
                }

                // == entity properties ==
                if (object.positionParent) {
                    // prepend a component that handles position
                    // reason is because it's not possible to know the dimension before all components are attached
                    componentInstances.unshift(new RelativePositioner({
                        offset: offset,
                        positionParent: object.positionParent,
                        positionSelf: object.positionSelf,
                        positionViewport: object.positionViewport
                    }));
                }
                if (object.scale) {
                    object.scale = toVector2(object.scale);
                }

                if (object.rotationDegrees) {
                    settingsObject.rotation = Utils.toRadian(object.rotationDegrees);
                }

                if (object.dimension) {
                    object.dimension = toRectangle(object.dimension);
                }
                if (object.boundingBox) {
                    object.boundingBox = toRectangle(object.boundingBox);
                }

                if (cachedModules[moduleName]) {
                    onGetConstructor(cachedModules[moduleName]);
                } else {
                    if (didPreload) {
                        // the developer had the intention to preload, yet this module was not found before: warn the developer
                        console.warn("WARNING: module " + moduleName + " was found but it was never preloaded.");
                    }
                    bento.require([moduleName], onGetConstructor);
                }
            };

            if (components && components.length) {
                Utils.forEach(components, function (item, i) {
                    if (item.module) {
                        // can be parsed as module (async)
                        toLoad += 1;
                        parse(item, function (instance) {
                            loaded += 1;
                            // add to instances
                            componentInstances[i] = instance;

                            // check if ready
                            if (didLoop && toLoad === loaded) {
                                onReadyToInstantiate();
                            }
                        });
                    } else {
                        // just add the item by itself
                        componentInstances[i] = item;
                    }
                });
            } else {
                // no components available
            }
            didLoop = true;

            if (toLoad === loaded) {
                onReadyToInstantiate();
            }
        };
        return uiLoader;
    };
    /**
     * Helper functions
     */
    var toVector2 = function toVector2(object) {
        return new Vector2(object.x || 0, object.y || 0);
    };
    var toRectangle = function toRectangle(object) {
        return new Rectangle(object.x || 0, object.y || 0, object.width || 0, object.height || 0);
    };
    /**
     * Helper module
     */
    var RelativePositioner = function RelativePositioner(settings) {
        var viewport = Bento.getViewport();
        //var position = toVector2(settings.position);

        var positionParent = toVector2(settings.positionParent);
        var positionSelf = toVector2(settings.positionSelf);
        var positionViewport = toVector2(settings.positionViewport);

        var offset = toVector2(settings.offset);
        //var type = settings.type;

        var entity;
        var component = {
            name: 'positioner',
            entity: null,
            start: function start() {
                window.addEventListener('resize', this.reposition, false);
                this.reposition();
            },
            destroy: function destroy() {
                window.removeEventListener('resize', this.reposition);
            },
            attached: function attached(data) {
                entity = data.entity;
            },
            reposition: function reposition() {

                entity.position.x = offset.x;
                entity.position.x += positionParent.x * (entity.parent ? entity.parent.dimension.width : 0); // /entity.scale.x;
                entity.position.x += positionSelf.x * entity.dimension.width * entity.scale.x;
                entity.position.x += positionViewport.x * viewport.width;

                entity.position.y = offset.y;
                entity.position.y += positionParent.y * (entity.parent ? entity.parent.dimension.height : 0); // /entity.scale.y;
                entity.position.y += positionSelf.y * entity.dimension.height * entity.scale.y;
                entity.position.y += positionViewport.y * viewport.height;

                //if ((type === 'relative' || type === 'parent') && entity && entity.parent) {
                //    // console.log(entity.parent.dimension, position, offset, entity.parent);
                //    // relative to parent
                //    entity.position.x = offset.x + position.x * entity.parent.dimension.width;
                //    entity.position.y = offset.y + position.y * entity.parent.dimension.height;
                //} else if (type === 'viewport') {
                //    // viewport
                //    entity.position.x = offset.x + position.x * viewport.width;
                //    entity.position.y = offset.y + position.y * viewport.height;
                //} else {
                //    // absolute
                //    entity.position.x = offset.x + position.x;
                //    entity.position.y = offset.y + position.y;
                //}
            }
        };
        return component;
    };

    /**
     * Preloading of modules: 
     * Requirement is that all ui modules need to be located in the folder structure 
     * `/assets/[assetgroup]/json/ui/`
     * This preloader will inspect all available json and cache-require all the modules
     * @snippet UILoader.preload|snippet
    UILoader.preload({
    onComplete: function (moduleCount) {}
    });
     */
    UILoader.preload = function (settings) {
        var onComplete = settings.onComplete;
        var allAssets = Bento.assets.getAssets();
        var allJson = allAssets.json;
        var modulesToLoad = {};
        var modulesCount = 0;
        var modulesLoaded = 0;
        var walk = function walk(json) {
            var moduleName = json.module;
            // never loaded?
            if (!cachedModules[moduleName]) {
                // using an object literal as set
                modulesToLoad[moduleName] = moduleName;
            }
            // loop through components
            Utils.forEach(json.components || [], function (component) {
                walk(component);
            });
        };
        var checkCompletion = function checkCompletion() {
            // count up and check completion
            modulesLoaded += 1;
            if (modulesLoaded === modulesCount && onComplete) {
                onComplete();
                onComplete = null;
            }
        };

        didPreload = true;

        Utils.forEach(allJson, function (json, assetName) {
            if (assetName.toLowerCase().indexOf('ui/') === 0) {
                // loop through json and collect all modules
                walk(json);
            }
        });

        // once we have collected all modules, we start loading them
        modulesCount = Utils.getKeyLength(modulesToLoad);
        if (modulesCount === 0) {
            // unless we didnt find any new modules
            onComplete(modulesLoaded);
            return;
        }
        Utils.forEach(modulesToLoad, function (moduleName) {
            // preload it
            bento.require([moduleName], function (Constructor) {
                // cache constructor
                cachedModules[moduleName] = Constructor;
                checkCompletion();
            });
        }, function (error) {
            Utils.log('ERROR: ' + error);
            checkCompletion();
        });
    };
    /**
     * Clear entire cache (good for bento quick reload...? would need to re-preload though)
     */
    UILoader.clearCache = function () {
        cachedModules = {};
    };
    return UILoader;
});