'use strict';

/**
 * Box2D wrapper module.
 * (Note: Compatibility with Canvas+ Box2D bindings has been dropped)
 * @moduleName Box2D
 *
 * Usage:
 * Add to your game's index.html:

    <script type="text/javascript" src="lib/box2dweb.js"></script>

 *
 * Most of the time, you want to use the wrapper functions provided by this module.
 * If you want to go under the hood, see the Box2DFlash docs:
 *   - http://www.box2dflash.org/docs/2.1a/reference/
 *
 * Also the LÖVE physics docs are helpful, but beware API differences:
 *   - https://love2d.org/wiki/love.physics
 *
 * Create a body:
 *
 * Handling collisions:
 *
 * Using sensors:
 * 
 * Limitations:
 * - There is only 1 physics world. If you call Box2D.setup() again, the world will be replaced.
 * - You must go under the hood to work with joints, or make on-the-fly changes to shapes and fixtures.
 *
 * @snippet Box2D.setup
Box2D.setup({
    gravity: new Vector2(0, Box2D.toPixels(9.8)),
    canSleep: true,
    fixedSpeed: true,
    debugDraw: false,
    debugLayer: 999999999999,
});
Bento.objects.attach(Box2D);

 *
 * @snippet Box2D.Body
Box2D.Body({
    type: 'dynamic',
    fixture: {
        shape: new Box2D.BoxShape(12, 14),
        friction: 1,
        restitution: 0.5,
        density: 0.3,
    }
})
 */
bento.define('modules/box2d', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/entity', 'bento/eventsystem', 'bento/utils', 'bento/packedimage'], function (Bento, Vector2, Rectangle, Sprite, Entity, EventSystem, Utils, PackedImage) {
    'use strict';

    var Box2D = window.Box2D;
    var b2Math = Box2D.Common.Math.b2Math;
    var b2Vec2 = Box2D.Common.Math.b2Vec2;

    var box2DWorld = void 0;
    var meterToPixels = 32; // 1 meter = 32
    var pixelsToMeter = 1 / meterToPixels;
    var gravity = new Box2D.Common.Math.b2Vec2(0, 9.8);
    var canSleep = true;
    var clearForces = true;
    var _isSimulating = false;
    var deferedFunctions = [];

    // when fixedSpeed is true, Box2D may step multiple times per frame to accommodate for different gameSpeed.
    var baseSpeed = 1.0; // = 60fps
    var fixedSpeed = true;
    var gainedSpeed = 0;
    var onStepFunctions = [];

    // distance / vector conversions
    var toPixels = function toPixels(m) {
        return meterToPixels * m;
    };
    var toMeters = function toMeters(px) {
        return pixelsToMeter * px;
    };
    var toBox2D = function toBox2D(vec2) {
        return new b2Vec2(pixelsToMeter * vec2.x, pixelsToMeter * vec2.y);
    };
    var toBento = function toBento(vec2) {
        return new Vector2(meterToPixels * vec2.x, meterToPixels * vec2.y);
    };

    // conversion without fixing units
    var toVector2 = function toVector2(vec2) {
        return new Vector2(vec2.x, vec2.y);
    };
    var toB2Vec2 = function toB2Vec2(vec2) {
        return new b2Vec2(vec2.x, vec2.y);
    };

    // enums
    var bodyTypeStringToEnum = {
        'static': Box2D.Dynamics.b2Body.b2_staticBody,
        'dynamic': Box2D.Dynamics.b2Body.b2_dynamicBody,
        'kinematic': Box2D.Dynamics.b2Body.b2_kinematicBody
    };
    var bodyTypeEnumToString = {};
    for (var k in bodyTypeStringToEnum) {
        var v = bodyTypeStringToEnum[k];
        bodyTypeEnumToString[v] = k;
    }

    // debug drawing
    var debugDraw = new Box2D.Dynamics.b2DebugDraw();
    var canvas = void 0;
    var context = void 0;
    var debugSprite = void 0;
    var debugEntity = void 0;
    var debugLayer = void 0;

    var module = {
        name: 'Box2D',
        z: 0,
        pxToM: pixelsToMeter,
        mToPx: meterToPixels,

        toPixels: toPixels,
        toMeters: toMeters,
        toVector2: toVector2,
        toB2Vec2: toB2Vec2,
        toBox2D: toBox2D,
        toBento: toBento,

        // Expose some commonly used internals
        // (as a rule of thumb, anything starting with 'b2' requires manual conversion of distance units and vector types)
        b2WorldManifold: Box2D.Collision.b2WorldManifold,
        b2Math: Box2D.Common.Math.b2Math,
        b2Vec2: Box2D.Common.Math.b2Vec2,
        b2CircleShape: Box2D.Collision.Shapes.b2CircleShape,
        b2EdgeChainDef: Box2D.Collision.Shapes.b2EdgeChainDef,
        b2MassData: Box2D.Collision.Shapes.b2MassData,
        b2PolygonShape: Box2D.Collision.Shapes.b2PolygonShape,
        b2Shape: Box2D.Collision.Shapes.b2Shape,
        b2DistanceJointDef: Box2D.Dynamics.Joints.b2DistanceJointDef,
        b2FrictionJointDef: Box2D.Dynamics.Joints.b2FrictionJointDef,
        b2GearJointDef: Box2D.Dynamics.Joints.b2GearJointDef,
        b2JointEdge: Box2D.Dynamics.Joints.b2JointEdge,
        b2LineJointDef: Box2D.Dynamics.Joints.b2LineJointDef,
        b2PrismaticJointDef: Box2D.Dynamics.Joints.b2PrismaticJointDef,
        b2PulleyJointDef: Box2D.Dynamics.Joints.b2PulleyJointDef,
        b2RevoluteJointDef: Box2D.Dynamics.Joints.b2RevoluteJointDef,
        b2WeldJointDef: Box2D.Dynamics.Joints.b2WeldJointDef,

        Box2D: Box2D,
        box2DWorld: box2DWorld,
        debugDraw: false,
        debugAlpha: 1,

        isPaused: false,

        /**
         * Is Box2d currently inside a simulation step?
         */
        isSimulating: function isSimulating() {
            return _isSimulating;
        },

        /**
         * Push a function that needs to be run after the simulation
         */
        defer: function defer(fn) {
            deferedFunctions.push(fn);
        },

        /**
         * Setup box2d world
         */
        setup: function setup(settings) {
            if (Utils.isDefined(settings.gravity)) gravity = toBox2D(settings.gravity);
            if (Utils.isDefined(settings.canSleep)) canSleep = settings.canSleep;
            if (Utils.isDefined(settings.z)) this.z = settings.z;
            if (Utils.isDefined(settings.meterToPixels)) meterToPixels = settings.meterToPixels;
            if (Utils.isDefined(settings.baseSpeed)) baseSpeed = settings.baseSpeed;
            if (Utils.isDefined(settings.fixedSpeed)) fixedSpeed = settings.fixedSpeed;
            pixelsToMeter = 1 / meterToPixels;
            this.pxToM = pixelsToMeter;
            this.mToPx = meterToPixels;
            this.debugDraw = settings.debugDraw;
            this.debugAlpha = settings.debugAlpha;
            debugLayer = settings.debugLayer;
            onStepFunctions = [];
            module.isPaused = false;
            // module.logAvailableApi();
        },
        start: function start(data) {
            // setup debug drawing
            var viewport = Bento.getViewport();
            var b2DebugDraw = Box2D.Dynamics.b2DebugDraw;

            // setup world
            box2DWorld = new Box2D.Dynamics.b2World(gravity, canSleep);
            box2DWorld.SetContinuousPhysics(true);
            this.box2DWorld = box2DWorld;

            // collision handlers

            // for any contact event, we check the fixture Settings object for collision handlers
            // if we find a handler, call it with the other fixture's Settings object as the first argument

            var listener = new Box2D.Dynamics.b2ContactListener();

            listener.BeginContact = function (contact) {
                var a = contact.GetFixtureA().GetUserData();
                var b = contact.GetFixtureB().GetUserData();
                if (a.onBeginContact) a.onBeginContact(b, contact);
                if (b.onBeginContact) b.onBeginContact(a, contact);
            };

            listener.EndContact = function (contact) {
                var a = contact.GetFixtureA().GetUserData();
                var b = contact.GetFixtureB().GetUserData();
                if (a.onEndContact) a.onEndContact(b, contact);
                if (b.onEndContact) b.onEndContact(a, contact);
            };
            listener.PostSolve = function (contact, impulse) {
                var a = contact.GetFixtureA().GetUserData();
                var b = contact.GetFixtureB().GetUserData();
                if (a.onPostSolve) a.onPostSolve(b, contact, impulse);
                if (b.onPostSolve) b.onPostSolve(a, contact, impulse);
            };
            listener.PreSolve = function (contact, oldManifold) {
                var a = contact.GetFixtureA().GetUserData();
                var b = contact.GetFixtureB().GetUserData();
                if (a.onPreSolve) a.onPreSolve(b, contact, oldManifold);
                if (b.onPreSolve) b.onPreSolve(a, contact, oldManifold);
            };
            box2DWorld.SetContactListener(listener);

            // debug drawing
            if (!this.debugDraw) {
                return;
            }
            canvas = document.createElement('canvas');
            canvas.width = viewport.width;
            canvas.height = viewport.height;
            context = canvas.getContext('2d');
            debugSprite = new Sprite({
                image: new PackedImage(canvas)
            });
            debugEntity = new Entity({
                z: debugLayer || 1000000000000,
                name: 'Box2DDebug',
                alpha: this.debugAlpha,
                components: [debugSprite]
                // float: true
            });
            debugDraw.SetSprite(context);
            debugDraw.SetDrawScale(meterToPixels);
            debugDraw.SetFlags(b2DebugDraw.e_shapeBit);
            box2DWorld.SetDebugDraw(debugDraw);
            Bento.objects.attach(debugEntity);
        },
        update: function update(data) {
            if (module.isPaused) {
                return;
            }
            var dt = baseSpeed / 60;

            if (fixedSpeed) {
                gainedSpeed += data.speed;
            } else {
                gainedSpeed = 1.0; // step once
                dt *= data.speed; // variable timestep
            }

            while (gainedSpeed > 0) {
                gainedSpeed -= 1.0;

                onStepFunctions.forEach(function (f) {
                    return f(dt);
                });

                // step box2d simulation
                _isSimulating = true;
                box2DWorld.Step(dt, 10, 10);

                if (clearForces) {
                    box2DWorld.ClearForces();
                }

                _isSimulating = false;
                // run defered functions
                Utils.forEach(deferedFunctions, function (fn) {
                    return fn();
                });
                // clean up
                deferedFunctions = [];
            }
        },
        draw: function draw(data) {
            if (this.debugDraw) {
                canvas.texture = null;
                box2DWorld.DrawDebugData();
            }
        },


        // if this works, could subtract from gainedSpeed so that the simulation goes unaffected?
        forceStep: function forceStep() {
            var dt = baseSpeed / 60;
            onStepFunctions.forEach(function (f) {
                return f(dt);
            });
            _isSimulating = true;
            box2DWorld.Step(dt, 10, 10);
            if (clearForces) {
                box2DWorld.ClearForces();
            }
            _isSimulating = false;
            Utils.forEach(deferedFunctions, function (fn) {
                return fn();
            });
            deferedFunctions = [];
        },
        addStepFunction: function addStepFunction(fn) {
            onStepFunctions.push(fn);
        },
        removeStepFunction: function removeStepFunction(fn) {
            var i = onStepFunctions.indexOf(fn);
            if (i !== -1) {
                onStepFunctions.splice(i, 1);
            }
        },
        isCircleShape: function isCircleShape(shape) {
            return shape instanceof Box2D.Collision.Shapes.b2CircleShape;
        },
        isPolygonShape: function isPolygonShape(shape) {
            return shape instanceof Box2D.Collision.Shapes.b2PolygonShape;
        },


        // shorthands - do we need wrapper classes for these?
        CircleShape: function CircleShape(radius, center) {
            var shape = new Box2D.Collision.Shapes.b2CircleShape(toMeters(radius));
            if (center) {
                shape.SetLocalPosition(toBox2D(center));
            }
            return shape;
        },
        BoxShape: function BoxShape(halfW, halfH, center) {
            var shape = new Box2D.Collision.Shapes.b2PolygonShape();
            halfW = toMeters(halfW);
            halfH = toMeters(halfH);
            var ox = center ? toMeters(center.x) : 0;
            var oy = center ? toMeters(center.y) : 0;
            if (center) {
                var points = [new b2Vec2(-halfW + ox, -halfH + oy), new b2Vec2(halfW + ox, -halfH + oy), new b2Vec2(halfW + ox, halfH + oy), new b2Vec2(-halfW + ox, halfH + oy)];
                shape.SetAsArray(points, points.length);
            } else {
                shape.SetAsBox(halfW, halfH);
            }
            return shape;
        },
        PolygonShape: function PolygonShape(points) {
            var len = points.length;
            var shape = new Box2D.Collision.Shapes.b2PolygonShape();
            for (var i = 0; i < len; i++) {
                var p = points[i];
                p.x = toMeters(p.x);
                p.y = toMeters(p.y);
            }
            shape.SetAsArray(points, len);
            return shape;
        },
        PolygonShapeRaw: function PolygonShapeRaw(points) {
            return Box2D.Collision.Shapes.b2PolygonShape.AsArray(points, points.length);
        },


        // utility for drawing shapes without debugdraw
        drawShape: function drawShape(renderer, shape) {
            var color = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [0, 0, 0, 1.0];

            switch (shape.GetType()) {
                case Box2D.Collision.Shapes.b2Shape.e_circleShape:
                    var p = shape.GetLocalPosition();
                    var r = shape.GetRadius();
                    renderer.fillCircle(color, toPixels(p.x), toPixels(p.y), toPixels(r));
                    break;
                case Box2D.Collision.Shapes.b2Shape.e_polygonShape:
                    var a = void 0,
                        b = void 0,
                        vertices = shape.GetVertices();
                    for (var i = 0; i < vertices.length - 1; i++) {
                        a = vertices[i];
                        b = vertices[i + 1];
                        renderer.drawLine(color, toPixels(a.x), toPixels(a.y), toPixels(b.x), toPixels(b.y), 1);
                    }
                    a = vertices[vertices.length - 1];
                    b = vertices[0];
                    renderer.drawLine(color, toPixels(a.x), toPixels(a.y), toPixels(b.x), toPixels(b.y), 1);
                    break;
                default:
                    break;
            }
        },


        /**
         * category:   what type is this fixture?
         * mask:       what types does this fixture collide with?
         * group:      two fixtures must have the same group in order to collide?
         * @snippet Box2D.Filter|snippet
        Box2D.Filter(${1:category}, ${2:mask}, ${3:group})
        */
        Filter: function Filter() {
            var category = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0x0001;
            var mask = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0xffff;
            var group = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;

            var filter = new Box2D.Dynamics.b2FilterData();
            filter.categoryBits = category;
            filter.maskBits = mask;
            filter.groupIndex = group;
            return filter;
        },


        // Wrapper component around b2Body -- attach this to an entity to hijack position and rotation
        Body: function Body(settings) {
            var entity = void 0;
            var box2DBody = void 0;
            var hasStarted = false;
            var onStep = null;

            var component = {
                name: 'body',

                fixtures: null,

                start: function start(data) {
                    var fixtures = void 0;
                    var bodyDef = new Box2D.Dynamics.b2BodyDef();
                    var setBodyDef = function setBodyDef(name) {
                        // sets body property if existant in settings
                        if (Utils.isDefined(settings[name])) {
                            bodyDef[name] = settings[name];
                        }
                    };
                    if (!entity.isAdded) {
                        return;
                    }

                    // body settings
                    bodyDef.type = settings.type ? bodyTypeStringToEnum[settings.type] : Box2D.Dynamics.b2Body.b2_staticBody;
                    bodyDef.position.Set(toMeters(entity.position.x), toMeters(entity.position.y));
                    bodyDef.angle = entity.rotation;
                    setBodyDef('active');
                    setBodyDef('allowSleep');
                    setBodyDef('angularDamping');
                    setBodyDef('angularVelocity');
                    setBodyDef('awake');
                    setBodyDef('bullet');
                    setBodyDef('fixedRotation');
                    setBodyDef('inertiaScale');
                    setBodyDef('linearDamping');
                    setBodyDef('linearVelocity');

                    box2DBody = box2DWorld.CreateBody(bodyDef);

                    if (box2DBody === null && module.isSimulating()) {
                        // TODO: defer start function automatically
                        console.error('Could not instantiate Body. Use Box2D.defer to instantiate a body outside the Box2D simulation step.');
                        return;
                    }

                    // fixture settings
                    fixtures = settings.fixture || settings.fixtures || [];
                    if (!Utils.isArray(fixtures)) {
                        fixtures = [fixtures];
                    }
                    Utils.forEach(fixtures, function (fixture, i, l, breakLoop) {
                        component.createFixture(fixture);
                    });

                    component.fixtures = fixtures;

                    if (settings.onStep) {
                        onStep = settings.onStep;
                        module.addStepFunction(onStep);
                    }

                    // userdata is reference to component
                    box2DBody.SetUserData(this);
                    hasStarted = true;
                },
                destroy: function destroy(data) {
                    var cleanup = function cleanup() {
                        box2DWorld.DestroyBody(box2DBody);
                        hasStarted = false;
                        if (onStep) {
                            module.removeStepFunction(onStep);
                        }
                    };
                    if (module.isSimulating()) {
                        // wait a moment until Box2D is done simulating
                        module.defer(cleanup);
                    } else {
                        // can safely remove
                        cleanup();
                    }
                },
                update: function update(data) {
                    // update transform to match rigid body
                    var pos = box2DBody.GetPosition();
                    entity.position.x = toPixels(pos.x);
                    entity.position.y = toPixels(pos.y);
                    if (!settings.ignoreRotation) {
                        entity.rotation = box2DBody.GetAngle();
                    }
                },
                attached: function attached(data) {
                    entity = data.entity;
                },
                getB2Body: function getB2Body() {
                    if (hasStarted === false) {
                        console.error('Body is not initialized yet. Wait until entity start');
                        return null;
                    }
                    return box2DBody;
                },


                // wrapper methods which add / remove fixtures using a 'settings' object
                createFixture: function createFixture(fixture) {
                    var fixtureDef = new Box2D.Dynamics.b2FixtureDef();
                    var setFixtureDef = function setFixtureDef(name) {
                        // sets fixture property if existant in settings
                        if (Utils.isDefined(fixture[name])) {
                            fixtureDef[name] = fixture[name];
                        }
                    };
                    fixtureDef.shape = fixture.shape || new Box2D.Collision.Shapes.b2CircleShape(toMeters(16));
                    setFixtureDef('friction');
                    setFixtureDef('restitution');
                    setFixtureDef('density');
                    setFixtureDef('isSensor');
                    setFixtureDef('filter');

                    // fixture settings act as user data, for use in collision handlers.
                    // body and entity fields are added to the settings for convenience.
                    // and we need to keep a reference to the underlying fixture in case we want to remove it later.
                    fixtureDef.userData = fixture;
                    fixture.body = component;
                    fixture.entity = entity;
                    fixture.box2DFixture = box2DBody.CreateFixture(fixtureDef);
                    if (component.fixtures) {
                        component.fixtures.push(fixture);
                    }
                },
                destroyFixture: function destroyFixture(fixture) {
                    var index = component.fixtures.indexOf(fixture);
                    if (index === -1) {
                        return;
                    }
                    box2DBody.DestroyFixture(fixture.box2DFixture);
                    component.fixtures.splice(index, 1);
                },


                // wrapper methods that automatically convert between Vector2 / b2Vec2

                // NOTE: there's so many of these...
                // it makes me wonder about what the performance overhead is like when creating closures?
                // My understanding is that prototypes would lead to smaller creation overhead and less memory usage?

                setPositionAndAngle: function setPositionAndAngle(pos, angle) {
                    box2DBody.setPositionAndAngle(toBox2D(pos), angle);
                },
                getAngle: function getAngle() {
                    return box2DBody.GetAngle();
                },
                setAngle: function setAngle(angle) {
                    box2DBody.SetAngle(angle);
                },
                getPosition: function getPosition() {
                    return toBento(box2DBody.GetPosition());
                },
                setPosition: function setPosition(pos) {
                    box2DBody.SetPosition(toBox2D(pos));
                },
                getWorldCenter: function getWorldCenter() {
                    return toBento(box2DBody.GetWorldCenter());
                },
                getLocalCenter: function getLocalCenter() {
                    return toBento(box2DBody.GetLocalCenter());
                },
                getLinearVelocity: function getLinearVelocity() {
                    return toBento(box2DBody.GetLinearVelocity());
                },
                setLinearVelocity: function setLinearVelocity(vec) {
                    box2DBody.SetLinearVelocity(toBox2D(vec));
                },
                getAngularVelocity: function getAngularVelocity() {
                    return box2DBody.GetAngularVelocity();
                },
                setAngularVelocity: function setAngularVelocity(omega) {
                    box2DBody.SetAngularVelocity(omega);
                },
                applyForce: function applyForce(force, point) {
                    box2DBody.ApplyForce(toBox2D(force), toBox2D(point));
                },


                // this function is missing from box2dweb so we implement it ourselves
                // applyAngularImpulse (impulse) {

                //     // impulse is in kg*m^2/s, so we need to scale down twice
                //     impulse = toMeters(toMeters(impulse));

                //     if (box2DBody.m_type !== Box2D.Dynamics.b2Body.b2_dynamicBody) {
                //         return;
                //     }
                //     if (!box2DBody.IsAwake()) {
                //         box2DBody.SetAwake(true);
                //     }
                //     box2DBody.m_angularVelocity += box2DBody.m_invI*impulse;
                // },
                applyTorque: function applyTorque(torque) {
                    box2DBody.ApplyTorque(toMeters(toMeters(torque)));
                },
                applyImpulse: function applyImpulse(impulse, point) {
                    box2DBody.ApplyImpulse(toBox2D(impulse), toBox2D(point));
                },

                // split(), merge()  // too low level for this wrapper?
                getMass: function getMass() {
                    return box2DBody.GetMass();
                },

                // getInertia () {
                //     let I = 0.0;
                //     component.fixtures.forEach(function (f) {
                //         let density = f.GetDensity();
                //         let shape = f.GetShape();
                //         if (density === 0.0) {
                //             return; // continue
                //         }
                //         switch (shape.GetType()) {
                //             case b2Shape.e_circleShape:
                //                 let radius = shape.GetRadius();
                //                 let mass = density * Math.PI * radius * radius;
                //                 // I += mass * (0.5 * radius * radius + )
                //                 break;
                //             case b2Shape.e_polygonShape:
                //                 break;
                //             default:
                //                 Utils.log('ERROR: unknown fixture type: '+shape.GetType());
                //         }
                //     });
                //     return toPixels(toPixels(box2DBody.GetInertia()));
                // },
                // getMassData(), setMassData(), resetMassData()
                getWorldPoint: function getWorldPoint(localPoint) {
                    return toBento(box2DBody.GetWorldPoint(toBox2D(localPoint)));
                },
                getWorldVector: function getWorldVector(localVector) {
                    // doesn't exist in cocoon so we implement it ourselves

                    // sanity check
                    if (Utils.isDev() && !Utils.isCocoonJs()) {
                        var a = toBento(box2DBody.GetWorldVector(toBox2D(localVector)));
                        var b = localVector.clone().rotateRadian(box2DBody.GetAngle());
                        if (Math.abs(a.x - b.x) > 0.00000001 || Math.abs(a.y - b.y) > 0.00000001) {
                            console.log('getWorldVector wrong result! expected:', a, 'actual:', b);
                        }
                    }

                    return localVector.clone().rotateRadian(box2DBody.GetAngle());
                },
                getLocalPoint: function getLocalPoint(worldPoint) {
                    return toBento(box2DBody.GetLocalPoint(toBox2D(worldPoint)));
                },
                getLocalVector: function getLocalVector(worldVector) {
                    return toBento(box2DBody.GetLocalVector(toBox2D(worldVector)));
                },
                getLinearVelocityFromWorldPoint: function getLinearVelocityFromWorldPoint(worldPoint) {
                    return toBento(box2DBody.GetLinearVelocityFromWorldPoint(toBox2D(worldPoint)));
                },
                getLinearVelocityFromLocalPoint: function getLinearVelocityFromLocalPoint(localPoint) {
                    return toBento(box2DBody.GetLinearVelocityFromLocalPoint(toBox2D(localPoint)));
                },
                getLinearDamping: function getLinearDamping() {
                    return box2DBody.GetLinearDamping();
                },
                setLinearDamping: function setLinearDamping(linearDamping) {
                    box2DBody.SetLinearDamping(linearDamping);
                },
                getAngularDamping: function getAngularDamping() {
                    return box2DBody.GetAngularDamping();
                },
                setAngularDamping: function setAngularDamping(angularDamping) {
                    box2DBody.SetAngularDamping(angularDamping);
                },
                getType: function getType() {
                    return bodyTypeEnumToString[box2DBody.GetType()];
                },
                setType: function setType(type) {
                    box2DBody.SetType(bodyTypeStringToEnum[type]);
                },
                isBullet: function isBullet() {
                    return box2DBody.IsBullet();
                },
                setBullet: function setBullet(flag) {
                    box2DBody.SetBullet(flag);
                },
                setSleepingAllowed: function setSleepingAllowed(flag) {
                    box2DBody.SetSleepingAllowed(flag);
                },
                isSleepingAllowed: function isSleepingAllowed() {
                    return box2DBody.IsSleepingAllowed();
                },
                setAwake: function setAwake(flag) {
                    box2DBody.SetAwake(flag);
                },
                isAwake: function isAwake() {
                    return box2DBody.IsAwake();
                },
                setFixedRotation: function setFixedRotation(flag) {
                    box2DBody.SetFixedRotation(flag);
                },
                isFixedRotation: function isFixedRotation() {
                    return box2DBody.IsFixedRotation();
                },
                setActive: function setActive(flag) {
                    box2DBody.SetActive(flag);
                },
                isActive: function isActive() {
                    return box2DBody.IsActive();
                },
                getFixtureList: function getFixtureList() {
                    return box2DBody.GetFixtureList();
                },
                getJointList: function getJointList() {
                    return box2DBody.GetJointList();
                },
                getControllerList: function getControllerList() {
                    return box2DBody.GetControllerList();
                },
                getContactList: function getContactList() {
                    return box2DBody.GetContactList();
                },
                getNext: function getNext() {
                    return box2DBody.GetNext();
                },


                // Userdata always refers to the Body wrapper component itself.
                // getUserData () {
                //     return box2DBody.GetUserData();
                // },
                // setUserData (data) {
                //     box2DBody.SetUserData(data);
                // },
                getWorld: function getWorld() {
                    return box2DWorld;
                }
            };

            return component;
        },


        // wrapper functions of the world itself:

        /**
         * For each collided fixture between p1 and p2, call fn with the following arguments:
         * fixture  : The fixture hit by the ray
         * point    : The point of initial intersection
         * normal   : The normal vector at the point of intersection
         * fraction : The fractional length along the ray of the intersection
         *
         * Callback can return the new length of the ray as a fraction of the original length.
         */
        rayCast: function rayCast(p1, p2, fn) {
            box2DWorld.RayCast(function (fixture, point, normal, fraction) {
                // let control = fn(fixture.GetUserData(), toBento(point), toBento(normal), fraction);

                // console.log({fixture});
                // pass b2vec because it's likely that we'll want to use that instead (e.g. for cutting shapes)
                var control = fn(fixture.GetUserData(), point, normal, fraction);
                return Utils.getDefault(control, 1);
            }, toBox2D(p1), toBox2D(p2));
        },
        queryShape: function queryShape(shape, transform, fn) {
            box2DWorld.QueryShape(function (fixture) {
                return fn(fixture.GetUserData());
            }, shape, transform);
        }
    };

    return module;
});