'use strict';

/**
 * As long as this module is attached, it will try to show a cloud load dialog every
 *  few seconds, but only if the cloud save has more progression than the local save.
 *
 * Likewise, you can perform a cloud save via this module, but only if
 *  the local save has progressed further than the cloud save.
 *
 * If the user declines the load dialog, it will stop showing automatically (but can )
 *
 * @moduleName CloudSaveChecker
 */
bento.define('modules/cloudsavechecker', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/entity', 'bento/eventsystem', 'bento/utils', 'bento/tween', 'modules/cloudsave', 'gui/cloudloaddialog', 'modules/savemanager'], function (Bento, Vector2, Rectangle, Entity, EventSystem, Utils, Tween, CloudSave, CloudLoadDialog, SaveManager) {
    'use strict';

    var timer = 0;
    var canAskLoad = false;

    var initialize = function initialize() {
        canAskLoad = true;
        EventSystem.on('saveToCloud', trySave);
    };

    var tryLoad = function tryLoad() {
        CloudSave.hasAnySave(function (success, snapshot, cloudSave) {
            console.log('>>> hasAnySave result (tryLoad):');
            console.log('>>> success:', JSON.stringify(success));
            // console.log('>>> snapshot:', JSON.stringify(snapshot));
            console.log('>>> localProgress:', SaveManager.load('cloudProgression'));
            console.log('>>> cloudProgress:', cloudSave && (cloudSave['cloudProgression'] || 0));
            if (success) {
                var cloudProgress = cloudSave['cloudProgression'] || 0;
                var localProgress = SaveManager.load('cloudProgression');
                if (localProgress < cloudProgress) {
                    // Allow player to load from cloud:
                    Bento.objects.attach(new CloudLoadDialog({
                        onSuccess: function onSuccess() {
                            // Allow the popup to show automatically next time.
                            SaveManager.save('cloudLoadDeclined', false);
                        }
                    }));
                } else {
                    // Better message would be something 'Current save is already up-to-date.'
                    EventSystem.fire('notification', { textId: 'cloudNoSaveFile' });
                }
                canAskLoad = false;
            } else {
                // No save found
                EventSystem.fire('notification', { textId: 'cloudNoSaveFile' });
            }
        });
    };

    // Similar to tryLoad, but silently fails and may reset the timer to try again if no save was found.
    var tryLoadAuto = function tryLoadAuto() {
        if (!canAskLoad) {
            return;
        }

        timer = 0;

        // If local save has less progression than cloud save, show cloud load dialog.

        CloudSave.hasAnySave(function (success, snapshot, cloudSave) {
            console.log('>>> hasAnySave result (tryLoadAuto):');
            console.log('>>> success:', JSON.stringify(success));
            // console.log('>>> snapshot:', JSON.stringify(snapshot));
            console.log('>>> localProgress:', SaveManager.load('cloudProgression'));
            console.log('>>> cloudProgress:', cloudSave && (cloudSave['cloudProgression'] || 0));
            if (success) {
                var cloudProgress = cloudSave['cloudProgression'] || 0;
                var localProgress = SaveManager.load('cloudProgression');
                if (localProgress < cloudProgress) {
                    // Allow player to load from cloud:
                    Bento.objects.attach(new CloudLoadDialog({
                        onDeclined: function onDeclined() {
                            SaveManager.save('cloudLoadDeclined', true);
                        }
                    }));
                } else {
                    // Do nothing, and stop checking.
                }
                canAskLoad = false;
            } else {
                if (!SaveManager.load('cloudLoadDeclined')) {
                    timer = 60 * 5; // allow module to try again in 5 seconds
                }
            }
        });
    };

    var trySave = function trySave() {

        var localProgress = SaveManager.load('cloudProgression');
        localProgress += 1;
        SaveManager.save('cloudProgression', localProgress);

        CloudSave.hasAnySave(function (success, snapshot, cloudSave) {
            console.log('>>> hasAnySave result (trySave):');
            console.log('>>> success:', JSON.stringify(success));
            // console.log('>>> snapshot:', JSON.stringify(snapshot));
            console.log('>>> localProgress:', localProgress);
            console.log('>>> cloudProgress:', cloudSave && (cloudSave['cloudProgression'] || 0));

            var shouldSave = void 0;
            if (success) {
                var cloudProgress = cloudSave['cloudProgression'] || 0;
                if (localProgress >= cloudProgress) {
                    shouldSave = true;
                    console.log('Saving to cloud, local save has more progression than existing cloudsave.');
                } else {
                    console.log('Not saving to cloud, local save has less progression than existing cloudsave.');
                }
            } else {
                if (snapshot === 'cloudSaveNone' || snapshot == null) {
                    shouldSave = true;
                    console.log('Saving to cloud, no cloudsave exists yet.');
                } else {
                    console.log('Could not check existing cloud save:', snapshot);
                }
            }
            if (shouldSave) {
                canAskLoad = false;
                CloudSave.saveToCloud(function (saveSuccess, reason) {
                    console.log('saveToCloud result:', saveSuccess, reason);
                });
            }
        });
    };

    // Show cloud load dialog immediately the next time it is attached.
    var quickStartEnabled = false;

    return {
        name: 'CloudSaveChecker',

        initialize: initialize,
        tryLoad: tryLoad,
        trySave: trySave,

        enableQuickStart: function enableQuickStart() {
            quickStartEnabled = true;
        },
        start: function start() {
            if (SaveManager.load('cloudLoadDeclined')) {
                timer = 0;
            } else {
                if (quickStartEnabled) {
                    quickStartEnabled = false;
                    timer = 10;
                } else {
                    timer = 60 * 3; // 3 seconds
                }
            }
        },
        update: function update() {
            if (timer > 0) {
                timer -= 1;
                if (timer <= 0) {
                    tryLoadAuto();
                }
            }
        }
    };
});