'use strict';

/**
 * Efficient particle system. 
 *
 * Limitations (for efficiency)
 * - No alpha  (you can put semi-transparent frames in the sheet if needed)
 * - All particles share the same spritesheet and anim (use multiple ParticleSystems if you need different sheets)
 * - Not framerate-independent (but ofc this shouldn't affect gameplay)
 *
 * @moduleName ParticleSystem
 * @snippet ParticleSystem.snippet
ParticleSystem({
    name: 'particleSystem',
    z: 0,
    imageName: '${1}',
    originRelative: new Vector2(${2:0.5}, ${3:0.5}),
    frameCountX: ${4:1},
    frameCountY: ${5:1},
    maxParticles: 256,
    animSpeed: 0.5,
    frames: [0,1,2,3] // if omitted, defaults to all frames in sheet
})
 */
bento.define('modules/particlesystem', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/entity', 'bento/utils'], function (Bento, Vector2, Rectangle, Entity, Utils) {
    'use strict';

    return function (settings) {
        var viewport = Bento.getViewport();

        var len = 0;
        var maxLen = settings.maxParticles;

        var particles = new Array(maxLen);
        for (var j = 0; j < maxLen; j++) {
            particles[j] = {
                life: 0,
                x: 0, y: 0,
                velX: 0, velY: 0,
                accX: 0, accY: 0,
                angle: 0, spin: 0,
                scale: 1, growth: 0,
                frame: 0, frameTimer: 0,
                sx: 0, sy: 0
            };
        }

        // image setup based on Sprite:

        var image;
        var frameWidth = settings.frameWidth || 0;
        var frameHeight = settings.frameHeight || 0;
        var frameCountX = settings.frameCountX || 1;
        var frameCountY = settings.frameCountY || 1;
        var ox = 0,
            oy = 0;

        if (settings.image) {
            image = settings.image;
        } else if (settings.imageName) {
            // load from string
            image = Bento.assets.getImage(settings.imageName);
        } else {
            Utils.log("ERROR: no image specified for ParticleSystem.");
            return;
        }
        if (!image) {
            Utils.log("ERROR: something went wrong with loading the image.");
            return;
        }
        // use frameWidth if specified (overrides frameCountX and frameCountY)
        if (frameWidth) {
            frameWidth = frameWidth;
            frameCountX = Math.floor(image.width / frameWidth);
        } else {
            frameCountX = frameCountX || 1;
            frameWidth = image.width / frameCountX;
        }
        if (frameHeight) {
            frameHeight = frameHeight;
            frameCountY = Math.floor(image.height / frameHeight);
        } else {
            frameCountY = frameCountY || 1;
            frameHeight = image.height / frameCountY;
        }

        if (settings.origin) {
            ox = settings.origin.x;
            oy = settings.origin.y;
        } else if (settings.originRelative) {
            ox = settings.originRelative.x * frameWidth;
            oy = settings.originRelative.y * frameHeight;
        }

        var animSpeed = Utils.isDefined(settings.animSpeed) ? settings.animSpeed : 0.5;
        var frames = settings.frames;
        var numFrames;

        if (frames) {
            numFrames = frames.length;
        } else {
            // default animation is all frames
            frames = [];
            numFrames = 0;
            for (var col = 0; col < frameCountX; col++) {
                for (var row = 0; row < frameCountY; row++) {
                    frames.push(numFrames++);
                }
            }
        }

        var particleSystem = {
            z: settings.z || 0,
            name: settings.name || 'particleSystem',
            start: function start(data) {},
            destroy: function destroy(data) {},

            getCount: function getCount() {
                return len;
            },

            add: function add(life, x, y, velX, velY, accX, accY, angle, spin, scale, growth, frameIndex) {

                if (len + 1 > maxLen) {
                    return;
                }
                var p = particles[len++];

                var frame = frames[frameIndex || 0];
                p.life = life;
                p.x = x;
                p.y = y;
                p.velX = velX || 0;
                p.velY = velY || 0;
                p.accX = accX || 0;
                p.accY = accY || 0;
                p.angle = angle || 0;
                p.spin = spin || 0;
                p.scale = scale == null ? 1 : scale;
                p.growth = growth || 0;
                p.frame = frameIndex || 0;
                p.frameTimer = 1;
                p.sx = frame % frameCountX * frameWidth;
                p.sy = Math.floor(frame / frameCountY) * frameHeight;
            },

            update: function update(data) {
                var i, p, velX, velY, life, frameTimer, frameIndex, frame;

                for (i = 0; i < len; i++) {
                    p = particles[i];
                    // check lifespan
                    if (p.life-- <= 0) {
                        // swap with the end particle, reduce length
                        len--;
                        var temp = particles[i];
                        particles[i] = particles[len];
                        particles[len] = temp;
                    }

                    // update transform / physics
                    velX = p.velX + p.accX;
                    velY = p.velY + p.accY;
                    p.velX = velX;
                    p.velY = velY;
                    p.x += velX;
                    p.y += velY;
                    p.angle += p.spin;
                    p.scale += p.growth;

                    frameTimer = p.frameTimer - animSpeed;

                    // advance to next frame
                    if (frameTimer <= 0.0) {
                        frameTimer += 1.0;
                        frameIndex = (p.frame + 1) % numFrames;
                        p.frame = frameIndex;
                        frame = frames[frameIndex];
                        p.sx = frame % frameCountX * frameWidth;
                        p.sy = Math.floor(frame / frameCountY) * frameHeight;
                    }
                    p.frameTimer = frameTimer;
                }
            },

            draw: function draw(data) {
                var renderer = data.renderer;
                var i, p, x, y, angle, scale;

                for (i = 0; i < len; i++) {
                    p = particles[i];
                    x = p.x;
                    y = p.y;
                    angle = p.angle;
                    scale = p.scale;
                    renderer.translate(x, y);
                    renderer.rotate(angle);
                    renderer.drawImage(image, p.sx, p.sy, frameWidth, frameHeight, -ox * scale, -oy * scale, frameWidth * scale, frameHeight * scale);

                    // more efficient than save/restore?
                    renderer.rotate(-angle);
                    renderer.translate(-x, -y);
                }
            }
        };
        return particleSystem;
    };
});