'use strict';

bento.define('modules/savemanager', ['bento', 'bento/utils'], function (Bento, Utils) {
    var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
        return typeof obj;
    } : function (obj) {
        return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };

    /**
     * Save keys
     */
    var Keys = {

        /**
         * Versions in the form 0xaabbcc, where aa = major, bb = minor, cc = patch
         */
        firstPlayedVersion: 0x000000, // inaccurate for versions before 1.0.3
        lastPlayedVersion: 0x000000, // useful for doing things such as save migration

        /**
         * Generic info for analytics
         */
        identifier: '', // user ID (can be device ID or googleplay/gameceter ID?)
        bootups: 0, // number of times the app was started
        firstBootDate: 0, // when was the app first launched
        retainedDays: {}, // collection of retention days
        levelsCleared: 0, // how many levels completed
        totalTimePlayed: 0, // approximately how long has the user played (for all sessions)

        /**
         * Notifications
         */
        notificationsAllowed: false, // can we do notifications?
        notificationsAsked: 0, // how many times have we asked for notifications?
        notifications: {
            // [name, defined by us]: [identifier (String), defined by Cocoon]
        },
        notificationsFreeGift: false,
        notifications7DayChallenge: false,

        /**
         * Time manager
         */
        timeOffset: null, // difference between user's time and server time
        lastKnownTime: 0, // last known server time
        initialStartingTime: 0, // the (server) time at which the player started playing this session

        /**
         * Game Specific Info
         */

        // When we press the Play button, which level do we go to?
        // Note: This could become invalid if the order of levels changed...
        //       Solution is to check if the level is actually unlocked, and recalculate if not
        nextLevel: -1,

        // increase every time the user clears a level (even if they already beat it before)
        winCount: 0,

        // real coin value.
        // if -1, coins will be calculated from stars earned
        // Note: coinCounter doesn't update when this is changed, you should fire 'updateCoins'
        coins: 999999999,

        // Number of hints that can be spent without having to watch an ad.
        hints: 0,

        'hasChar-1': true, // default character unlock (see hasChar base key)
        'hasScissor-1-1': true, // default scissor unlock (see hasScissor base key)
        'hasTheme-wave-blue': true, // default theme unlock (see hasTheme base key)

        character: '1', // which character did the player select?
        scissor: '1-1', // which scissor did the player select?
        theme: 'wave-blue', // which theme did the player select?

        scheduledFreeGift: 0, // time at which the next free gift will be available
        claimedFreeGifts: 0, // total number of free gifts claimed by the player

        /**
         * 7-day challenge
         */
        scheduledChallenge: 0, // when will the next 7-day challenge level become available?
        shownChallengePopup: false, // true if the player has seen the popup

        // integer value.
        // -1: player just started the game or claimed a chest
        // 0: player arrived at first segment
        // 4: player arrived at final segment (the one with the chest)
        // 5+: player didn't watch the ad so they are still at the chest!
        chestProgress: -1,

        // how many times did the player spin the wheel
        // (this is incremented as soon as the wheel starts spinning)
        wheelTimes: 0,

        /**
         * IAPs
         */
        noAds: false,

        /**
         * VIP
         */
        lastTimeVipClaimed: 0,
        VipFirstTimeClaim: false,
        VipLastValidated: 0,
        VipFailedValidations: 0,
        VipDowngraded: false,
        VIP: false,

        /**
         * Ads
         */
        totalInterstitials: 0,
        totalRewarded: 0,

        /**
         * iCloud Saves
         */
        cloudProgression: 0,
        cloudLoadDeclined: false
    };

    /**
     * Base keys. These allow to set up constructed SaveKeys.
     * For example, instead of saving highscores for each level,
     * we can create a new save key for each level, like 'highScores-lv1'
     */
    var Base = {
        // how many stars earned in each level.
        // e.g. 'stars-levelA1'
        // -1 = level never beaten.
        //  0 = level was most likely skipped, or was just recently unlocked
        stars: -1,

        // true if the player has unlocked a given character skin
        // e.g. 'hasChar-1'
        hasChar: false,

        // true if the player has unlocked a given scissor style
        // typically these are named 'x-y' where x = base style, y = recolor number
        // e.g. hasScissor-1-1
        hasScissor: false,

        // true if the player has unlocked a given theme
        // e.g. hasTheme-wave-pink
        hasTheme: false,

        // true if the player followed us on twitter, liked us on facebook, etc.
        claimedSocial: false
    };

    /**
     * Private functions, ignore these
     */
    var getParsedDefault = function getParsedDefault(def) {
        if (def !== null) {
            if (def.constructor === Array) {
                def = [].concat(def);
            } else if ((typeof def === 'undefined' ? 'undefined' : _typeof(def)) === 'object') {
                def = JSON.parse(JSON.stringify(def));
            }
        }
        return def;
    };

    var checkKey = function checkKey(key) {
        var base;
        var def;
        if (Keys.hasOwnProperty(key)) {
            return true;
        } else {
            //check if this is a constructed key
            if (key.indexOf('-') !== -1) {
                base = key.split('-')[0];
                if (key === base + '-') {
                    Utils.log('only base part of key provided!');
                    return false;
                }
                if (Base.hasOwnProperty(base)) {
                    //we now know this to be a valid (constructed) key, so add it to the Keys list
                    def = getParsedDefault(Base[base]);
                    Keys[key] = def;
                    return true;
                }
            }
            Utils.log('WARNING: save key ' + key + ' is not defined!');
            return false;
        }
    };

    /**
     * Public functions
     */
    var manager = {
        Keys: Keys,
        Base: Base,
        save: function save(key, value) {
            if (!checkKey(key)) {
                return;
            }
            Bento.saveState.save(key, value);
        },
        load: function load(key) {
            if (!checkKey(key)) {
                return undefined;
            }
            var def = getParsedDefault(Keys[key]);

            return Bento.saveState.load(key, def);
        },
        add: function add(key, value) {
            var current = manager.load(key);
            current += value;
            manager.save(key, current);
        }
    };

    return manager;
});