'use strict';

bento.define('modules/textstyles', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/entity', 'bento/eventsystem', 'bento/utils', 'bento/tween', 'modules/localization'], function (Bento, Vector2, Rectangle, Entity, EventSystem, Utils, Tween, Localization) {
    /**
     * Text styles manager
     * @moduleName TextStyles
     *
     * Snippet below is for creating a text object that uses this module:
     * @snippet Text|TextStyles
    Text({
        text: '$1',
        position: new Vector2(0, 0),
        scale: new Vector2(0.5, 0.5),
        fontSettings: TextStyles.get('${2:main}'),
        align: 'center',
        maxWidth: 64,
        maxHeight: 32,
        drawDebug: false
    });
     */

    var TextStyles = {

        /**
         * @snippet TextStyles.get()|Standard
            TextStyles.get('${1:standard}')
         */
        get: function get(styleName) {
            var style = {};
            if (TextStyles.hasOwnProperty(styleName)) {
                style = Utils.copyObject(TextStyles[styleName]);
            }
            if (style.hasOwnProperty('font') && fonts.hasOwnProperty(style.font)) {

                //get language font
                var font;
                if (fonts[style.font].hasOwnProperty(Localization.getLanguage())) {
                    font = fonts[style.font][Localization.getLanguage()];
                } else {
                    font = fonts[style.font].default;
                }
                style.font = font.fontName;

                //scale fontsize according to font settings
                if (font.hasOwnProperty('scale') && font.scale !== 1 && style.hasOwnProperty('fontSize')) {
                    style.fontSize = Math.round(font.scale * style.fontSize);
                }
                //offset font according to font settings
                if (font.offset.x !== 0 || font.offset.y !== 0) {
                    style.offset = Utils.copyObject(font.offset);
                    //scale by font size
                    style.offset.x = style.offset.x * (style.fontSize / 8);
                    style.offset.y = style.offset.y * (style.fontSize / 8);
                }
            }
            return style;
        }
    };
    var cascadeStyle = function cascadeStyle(style1, style2) {
        return Utils.extend(Utils.copyObject(style1), style2);
    };

    /**
     * FONTS
     */
    var JuaRegular = {
        fontName: 'Jua-Regular',
        offset: { x: 0, y: 0 },
        scale: 1
    };
    var MPLUSRounded1cBold = {
        fontName: 'MPLUSRounded1c-Bold',
        offset: { x: 0, y: 0 },
        scale: 1
    };
    var LuckiestGuy = {
        fontName: 'luckiestguy-regular',
        offset: { x: 0, y: 0 },
        scale: 1
    };
    var fonts = {
        main: {
            default: JuaRegular,
            ko: JuaRegular,
            de: MPLUSRounded1cBold,
            es: MPLUSRounded1cBold,
            fr: MPLUSRounded1cBold,
            it: MPLUSRounded1cBold,
            pt: MPLUSRounded1cBold,
            ru: MPLUSRounded1cBold,
            ja: MPLUSRounded1cBold,
            zh: MPLUSRounded1cBold,
            zt: MPLUSRounded1cBold
        },
        luckiest: {
            default: LuckiestGuy,
            ko: JuaRegular,
            de: MPLUSRounded1cBold,
            es: MPLUSRounded1cBold,
            fr: MPLUSRounded1cBold,
            it: MPLUSRounded1cBold,
            pt: MPLUSRounded1cBold,
            ru: MPLUSRounded1cBold,
            ja: MPLUSRounded1cBold,
            zh: MPLUSRounded1cBold,
            zt: MPLUSRounded1cBold
        }
    };

    /**
     * STYLES
     */
    TextStyles.main = {
        font: 'main',
        fontSize: 20,
        fontColor: '#000000',
        textBaseline: 'middle'
    };
    TextStyles.mainWhite = {
        font: 'main',
        fontSize: 20,
        fontColor: '#ffffff',
        textBaseline: 'middle'
    };
    TextStyles.mainVip = {
        font: 'main',
        fontSize: 20,
        fontColor: '#ffd235',
        textBaseline: 'middle'
    };
    TextStyles.loading = {
        font: 'main',
        fontSize: 20,
        fontColor: '#000000',
        textBaseline: 'bottom'
    };
    TextStyles.settingsButton = {
        font: 'main',
        fontSize: 11,
        fontColor: '#000000',
        textBaseline: 'middle'
    };
    TextStyles.heading = {
        font: 'main',
        fontSize: 24,
        fontColor: '#5333ff', // 'CLEAR' blue
        textBaseline: 'middle',
        lineWidth: 1.5,
        strokeStyle: '#ffffff'
    };
    TextStyles.bossFight = {
        font: 'main',
        fontSize: 24,
        fontColor: '#f44d4d', // 'CLEAR' blue
        textBaseline: 'middle',
        lineWidth: 1.5,
        strokeStyle: '#ffffff'
    };
    TextStyles.hint = {
        font: 'main',
        fontSize: 20,
        fontColor: '#f73b3b',
        textBaseline: 'middle',
        lineWidth: 1.5,
        strokeStyle: '#ffffff'
    };
    TextStyles.levelName = {
        font: 'main',
        fontSize: 20,
        fontColor: '#000000',
        textBaseline: 'middle',
        pixelStroke: true, // workaround for Cocoon bug
        lineWidth: 0.6,
        strokeStyle: '#ffffff'
    };
    TextStyles.retrySmall = {
        font: 'main',
        fontSize: 20,
        fontColor: '#000000',
        textBaseline: 'middle',
        pixelStroke: true,
        lineWidth: 0.85,
        strokeStyle: '#ffffff'
    };
    TextStyles.buttonLabel = {
        font: 'main',
        fontSize: 20,
        fontColor: '#ffffff',
        textBaseline: 'middle'
    };
    TextStyles.dev = {
        font: 'main',
        fontSize: 16,
        fontColor: '#ffffff',
        textBaseline: 'middle'
    };
    TextStyles.dialogTitle = {
        font: 'main',
        fontSize: 30,
        fontColor: '#000000',
        textBaseline: 'middle'
    };
    TextStyles.dialogText = {
        font: 'main',
        fontSize: 8,
        fontColor: '#000000',
        textBaseline: 'middle'
    };

    // for levelselect
    // should be replaced by counter to save resources?
    TextStyles.levelNumber = {
        font: 'main',
        fontSize: 20,
        fontColor: '#ffffff',
        textBaseline: 'middle'
    };

    TextStyles.vipTitle = {
        font: 'main',
        fontSize: 20,
        fontColor: '#000000',
        textBaseline: 'middle',
        pixelStroke: true, // workaround for Cocoon bug
        lineWidth: 0.6,
        strokeStyle: '#ffea87'
    };

    TextStyles.vipLabel = {
        font: 'main',
        fontSize: 20,
        fontColor: '#ffd235',
        textBaseline: 'top'
    };

    TextStyles.iapOverlay = {
        font: 'main',
        fontSize: 14,
        fontColor: '#ffffff',
        textBaseline: 'middle'
    };

    TextStyles.vipWindowTitle1 = {
        font: 'main',
        fontSize: 24,
        fontColor: '#ffffff',
        textBaseline: 'middle'
    };
    TextStyles.vipWindowTitle2 = {
        font: 'main',
        fontSize: 24,
        fontColor: '#ffd146',
        textBaseline: 'middle'
    };
    TextStyles.vipChars = {
        font: 'luckiest',
        fontSize: 24,
        fontColor: '#ffd146',
        textBaseline: 'middle'
    };
    TextStyles.vipMain = {
        font: 'main',
        fontSize: 8,
        fontColor: '#ffffff',
        textBaseline: 'top'
    };
    TextStyles.buttonSubscribeText1 = {
        font: 'main',
        fontSize: 14,
        fontColor: '#dd2b6a',
        textBaseline: 'middle'
    };
    TextStyles.buttonSubscribeText2 = {
        font: 'main',
        fontSize: 14,
        fontColor: '#2b54bc',
        textBaseline: 'middle'
    };
    TextStyles.vipTerms = {
        font: 'Helvetica',
        fontSize: 14,
        fontColor: '#aaaaaa',
        textBaseline: 'middle'
    };
    TextStyles.vipLink = {
        font: 'Helvetica',
        fontSize: 12,
        fontColor: '#28aeff',
        textBaseline: 'middle'
    };
    TextStyles.vipBannerTitle = {
        font: 'main',
        fontSize: 24,
        fontColor: '#ffd235',
        textBaseline: 'middle'
    };

    TextStyles.credits = {
        font: 'main',
        fontSize: 20,
        fontColor: '#ffffff',
        textBaseline: 'middle'
    };

    return TextStyles;
});