'use strict';

bento.define('screens/splash', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/components/clickable', 'bento/components/fill', 'bento/gui/text', 'bento/entity', 'bento/eventsystem', 'bento/utils', 'bento/screen', 'bento/tween', 'entities/luckykatlogo', 'globals', 'modules/tapticengine', 'modules/admanager'], function (Bento, Vector2, Rectangle, Sprite, Clickable, Fill, Text, Entity, EventSystem, Utils, Screen, Tween, LuckyKat, Globals, TapticEngine, AdManager) {

    var PIXI = window.PIXI;
    var earcut = window.earcut;

    var Terrain = function Terrain(settings) {
        var polygon = settings.polygon;


        var center = polygon.reduce(function (acc, v) {
            return acc.add(v);
        }).scalarMultiply(1 / polygon.length);

        var initialPos = center.clone();

        var pixiGraphic = void 0;

        var repaint = function repaint() {
            pixiGraphic = new PIXI.Graphics();
            var pixiPolygon = new PIXI.Polygon(polygon.map(function (v) {
                return new PIXI.Point(v.x - center.x, v.y - center.y);
            }));
            var scale = 0.5;
            var angle = 0;
            var texture = Utils.newTexture('texture/slice-white');

            // create mesh for inner fill
            var vertices = new Float32Array(pixiPolygon.points);
            var indices = new Uint16Array(earcut(vertices));
            var uvs = new Float32Array(vertices);
            for (var i = 0; i < vertices.length; i += 2) {
                var vec = new Vector2(uvs[i], uvs[i + 1]);
                if (angle) {
                    vec.rotateRadian(angle);
                }
                uvs[i] = (vec.x + initialPos.x) / (texture.width * scale);
                uvs[i + 1] = (vec.y + initialPos.y) / (texture.height * scale);
            }
            var mesh = new PIXI.mesh.Mesh(texture, vertices, uvs, indices, PIXI.mesh.Mesh.DRAW_MODES.TRIANGLES);
            pixiGraphic.addChild(mesh);
        };

        var behavior = {
            name: 'behavior',
            start: function start() {
                repaint();
            },
            update: function update() {},
            draw: function draw(_ref) {
                var renderer = _ref.renderer;

                renderer.drawPixi(pixiGraphic);
            }
        };
        var entity = new Entity({
            z: 0,
            name: 'splashTerrain',
            position: center.clone(),
            components: [behavior]
        });
        return entity;
    };

    var AnimScissor = function AnimScissor(settings) {
        var from = settings.from,
            to = settings.to,
            ease = settings.ease;


        var sprite = new Sprite({
            imageName: 'scissors/scissors-1-1',
            frameCountX: 2,
            originRelative: new Vector2(0.5, 0.5),
            animations: {
                default: { speed: 0.14, frames: [1, 0, 1], loop: false }
            }
        });

        var entity = new Entity({
            z: 0,
            name: 'animScissor',
            position: from.clone(),
            scale: new Vector2(0.14, 0.14),
            rotation: from.angleBetween(to) + Math.PI * 0.5,
            components: [sprite]
        });

        // move from a to b
        Tween({
            from: from, to: to,
            delay: 0,
            in: 18,
            ease: ease,
            onUpdate: function onUpdate(v, t) {
                v.copyInto(entity.position);
            }
        });
        Utils.timeout(4, function () {
            Bento.audio.playSound('sfx_scissor');
            TapticEngine.impact({ type: 'medium' });
        });

        return entity;
    };

    var onShow = function onShow(settings) {
        var loadDisplay = settings.loadDisplay;


        var viewport = Bento.getViewport(true);

        var bgScale = 0.16;
        var bgSprite = new PIXI.extras.TilingSprite(Utils.newTexture('texture/solid-wave-blue'), viewport.width / bgScale, viewport.height / bgScale);
        var bg = new Entity({
            z: 0,
            name: 'bg',
            float: true,
            position: new Vector2(0, 0),
            scale: new Vector2(bgScale, bgScale),
            components: [{
                name: 'behavior',
                draw: function draw(_ref2) {
                    var renderer = _ref2.renderer;

                    renderer.drawPixi(bgSprite);
                }
            }]
        });
        Bento.objects.attach(bg);

        var fadeSprite = new PIXI.Sprite(Utils.newTexture('texture/black', PIXI.SCALE_MODES.NEAREST));
        fadeSprite.width = viewport.width;
        fadeSprite.height = viewport.height;
        var fade = new Entity({
            z: 99999,
            name: 'fade',
            float: true,
            global: true,
            components: {
                name: 'behavior',
                draw: function draw(_ref3) {
                    var renderer = _ref3.renderer;

                    renderer.drawPixi(fadeSprite);
                }
            }
        });

        var isLeaving = false;
        var nextScreen = function nextScreen() {
            EventSystem.off('pointerDown', nextScreen);
            if (!isLeaving) {
                isLeaving = true;
                new Tween({
                    from: 0, to: 1, in: 12,
                    ease: 'easeInOutQuad',
                    onStart: function onStart() {
                        Bento.objects.attach(fade);
                    },
                    onUpdate: function onUpdate(v, t) {
                        fade.alpha = v;
                    },
                    onComplete: function onComplete() {
                        Bento.screens.show('screens/titlescreen');
                        Utils.timeout(8, function () {
                            Bento.audio.playMusic('bgm_loop_full');
                        });
                        var tweenEntity = new Entity({
                            z: 0,
                            name: 'fadeTween',
                            updateWhenPaused: 999,
                            ignoreGameSpeed: true,
                            components: [new Tween.TweenBehavior({
                                from: 1, to: 0, in: 16,
                                ease: 'easeInOutQuad',
                                onUpdate: function onUpdate(v, t) {
                                    fade.alpha = v;
                                },
                                onComplete: function onComplete() {
                                    Bento.objects.remove(fade);
                                }
                            })]
                        });
                        Bento.objects.attach(tweenEntity);

                        AdManager.showBanner();
                    }
                });
            }
        };

        EventSystem.on('pointerDown', nextScreen);

        var luckyKatName = new Entity({
            z: 0,
            name: 'luckyKatName',
            position: new Vector2(viewport.width * 0.5, viewport.height * 0.5 + 50),
            scale: new Vector2(0.16, 0.16),
            components: [new Sprite({
                imageName: 'luckykat-name',
                originRelative: new Vector2(0.5, 0.9)
            })]
        });
        var luckyKatLogo = new Entity({
            z: 0,
            name: 'luckyKatLogo',
            position: new Vector2(viewport.width * 0.5, viewport.height * 0.5 + 20),
            scale: new Vector2(0.16, 0.16),
            components: [new Sprite({
                imageName: 'luckykat-cat',
                originRelative: new Vector2(0.5, 0.9)
            })]
        });
        Bento.objects.attach(luckyKatName);
        Bento.objects.attach(luckyKatLogo);

        var doMeow = function doMeow() {
            var baseScale = luckyKatLogo.scale.clone();
            Tween({
                delay: 0,
                from: 0.0, to: 1.0, in: 26,
                ease: 'linear',
                onStart: function onStart() {
                    Bento.audio.setVolume(0.6, 'sfx_meow');
                    Bento.audio.playSound('sfx_meow');
                },
                onUpdate: function onUpdate(v, t) {
                    luckyKatLogo.rotation = Math.sin(t / 30 * Math.PI * 4) * 0.10 * (1 - v);
                    luckyKatLogo.scale.x = baseScale.x * (1.0 - Math.sin(v * Math.PI) * 0.1);
                    luckyKatLogo.scale.y = baseScale.y * (1.0 + Math.sin(v * Math.PI) * 0.1);
                }
            });
        };

        var p1 = new Vector2(viewport.width * 0.00 - 26, viewport.height * -0.05);
        var p2 = new Vector2(viewport.width * 1.00 + 26, viewport.height * -0.05);
        var p3 = new Vector2(viewport.width * 1.00 + 26, viewport.height * 0.38);
        var p4 = new Vector2(viewport.width * 0.00 - 26, viewport.height * 0.26);
        var p5 = new Vector2(viewport.width * 1.00 + 26, viewport.height * 0.59);
        var p6 = new Vector2(viewport.width * 0.00 - 26, viewport.height * 0.82);
        var p7 = new Vector2(viewport.width * 1.00 + 26, viewport.height * 1.00);
        var p8 = new Vector2(viewport.width * 0.00 - 26, viewport.height * 1.00);

        var topPart = new Terrain({ polygon: [p1, p2, p3, p4] });
        var middlePart = new Terrain({ polygon: [p4, p3, p5, p6] });
        var bottomPart = new Terrain({ polygon: [p6, p5, p7, p8] });

        Bento.objects.attach(topPart);
        Bento.objects.attach(middlePart);
        Bento.objects.attach(bottomPart);

        loadDisplay.position = middlePart.toLocalPosition(loadDisplay.position);
        middlePart.attach(loadDisplay);

        Utils.timeout(10, function () {
            var scis1 = new AnimScissor({ from: p4, to: p3, ease: 'linear' });
            Bento.objects.attach(scis1);
        });
        Utils.timeout(34, function () {
            var scis2 = new AnimScissor({ from: p5, to: p6, ease: 'linear' });
            Bento.objects.attach(scis2);
        });
        Utils.timeout(52, function () {
            var FallBehavior = function FallBehavior(_ref4) {
                var accel = _ref4.accel,
                    spin = _ref4.spin;

                var entity = void 0;
                var vel = new Vector2(0, 0);
                var rotvel = 0;
                return {
                    name: 'fallBehavior',
                    attached: function attached(data) {
                        entity = data.entity;
                    },
                    update: function update(_ref5) {
                        var speed = _ref5.speed;

                        vel.addTo(accel);
                        entity.position.addTo(vel);
                        rotvel = Utils.clamp(-0.1, rotvel + spin, 0.1);
                        entity.rotation += rotvel;
                    }
                };
            };
            bottomPart.attach(FallBehavior({
                accel: new Vector2(0, 0.38),
                spin: 0.0005
            }));
            Utils.timeout(5, function () {
                middlePart.attach(FallBehavior({
                    accel: new Vector2(0, 0.36),
                    spin: -0.0005
                }));
            });
            Utils.timeout(10, function () {
                topPart.attach(FallBehavior({
                    accel: new Vector2(0, 0.34),
                    spin: 0.0003
                }));
            });
            Utils.timeout(50, doMeow);
            Utils.timeout(94, nextScreen);
        });
    };

    var onHide = function onHide() {};

    return new Screen({
        onShow: onShow,
        onHide: onHide
    });
});