'use strict';

bento.define('utils', ['bento', 'bento/math/vector2', 'bento/math/rectangle', 'bento/components/sprite', 'bento/components/clickable', 'bento/entity', 'bento/utils', 'bento/tween', 'bento/eventsystem', 'modules/appavailability'], function (Bento, Vector2, Rectangle, Sprite, Clickable, Entity, Utils, Tween, EventSystem, AppAvailability) {
    /**
     * Extends utils with more useful functions.
     * See bento/utils
     */
    Utils.extend(Utils, {

        // Globals that should be preserved between hot reloads.
        currentLevel: 0,
        currentLevelGroup: 'default', // default = main level listing. Non-default groups may result in different UI being displayed (e.g. 7-day challenge)
        assetsLoaded: false,

        // Velcro framework uses this value, but Number.EPSILON is much smaller.
        // Not sure how significant this is.
        EPSILON: 1.192092896e-07,

        versionToHex: function versionToHex(versionStr) {
            var parts = versionStr.split('.').map(function (s) {
                return parseInt(s) & 0xff;
            });
            return parts[0] << 16 | parts[1] << 8 | parts[2];
        },


        // lambda urls
        urls: {
            'getDate': 'https://ermngnll7d.execute-api.us-east-1.amazonaws.com/latest/getdate'
        },

        // social urls
        urlSchemes: {
            'twitter': 'twitter://user?screen_name=LuckyKatStudios',
            'facebook': 'fb://profile/466268083523876',
            'instagram': 'instagram://user?username=luckykatstudios'
        },
        links: {
            'twitter': 'https://twitter.com/LuckyKatStudios',
            'facebook': 'https://www.facebook.com/luckykatstudios',
            'instagram': 'https://www.instagram.com/luckykatstudios/'
        },

        getStoreLink: function getStoreLink(device) {
            // TODO: double check these URLs before launch
            var links = {
                ios: 'https://itunes.apple.com/us/app/go-slice/id1441767098',
                android: 'https://play.google.com/store/apps/details?id=com.luckykat.slice'
            };
            var link = links.web;
            if (!device) {
                // try to determine automatically
                if (Utils.isNativeIos()) {
                    link = links['ios'];
                } else if (Utils.isNativeAndroid()) {
                    link = links['android'];
                }
            } else if (links[device]) {
                link = links[device];
            }
            return link;
        },


        // Open URL in new window
        openUrl: function openUrl(url) {
            if (window.cordova && window.cordova.InAppBrowser) {
                window.cordova.InAppBrowser.open(url, '_system');
            } else {
                window.open(url, '_blank');
            }
        },

        /**
         * Checks if url can be opened, always true for android
         * Uses https://github.com/ohh2ahh/AppAvailability
         */
        canOpenUrl: function canOpenUrl(name) {
            if (Utils.isNativeIos() && AppAvailability) {
                return AppAvailability[name] || false;
            } else {
                return true;
            }
        },

        isFB: function isFB() {
            return window.FBInstant;
        },
        isTablet: function isTablet() {
            // https://stackoverflow.com/a/17230890
            var screen = window.screen;
            var agent = navigator.userAgent;
            var isWebkit = agent.indexOf("AppleWebKit") > 0;
            var isIPad = agent.indexOf("iPad") > 0;
            var isIOS = agent.indexOf("iPhone") > 0 || agent.indexOf("iPod") > 0;
            var isAndroid = agent.indexOf("Android") > 0;
            var isNewBlackBerry = agent.indexOf("AppleWebKit") > 0 && agent.indexOf("BlackBerry") > 0;
            var isWebOS = agent.indexOf("webOS") > 0;
            var isWindowsMobile = agent.indexOf("IEMobile") > 0;
            var isSmallScreen = screen.width < 767 || isAndroid && screen.width < 1000;
            var isUnknownMobile = isWebkit && isSmallScreen;
            var isMobile = isIOS || isAndroid || isNewBlackBerry || isWebOS || isWindowsMobile || isUnknownMobile;
            var isTablet = isIPad || isMobile && !isSmallScreen;
            return isTablet;
        },
        timeout: function timeout(time, callback) {
            return Tween({
                from: 0,
                to: 1,
                in: time,
                ease: 'linear',
                onComplete: callback
            });
        },
        prompt: function prompt(title, text, callback) {
            if (Utils.isCocoonJS() && window.Cocoon) {
                window.Cocoon.Dialog.prompt({
                    title: title,
                    message: text
                }, {
                    success: callback,
                    cancel: function cancel() {}
                });
            } else {
                var res = window.prompt(text);
                if (res) {
                    callback(res);
                }
            }
        },
        deltaAngle: function deltaAngle(a, b) {
            while (b > a + Math.PI) {
                b -= 2 * Math.PI;
            }
            while (b < a - Math.PI) {
                b += 2 * Math.PI;
            }
            return b - a;
        },
        color: function color(string) {
            return string.match(/\w\w/g).map(function (x) {
                return parseInt(x, 16) / 255;
            });
        },
        reverseDictionary: function reverseDictionary(obj) {
            var res = {};
            for (var k in obj) {
                var v = obj[k];
                res[v] = k;
            }
            return res;
        },
        limitVector: function limitVector(vec, maxMagnitude) {
            var magSq = vec.sqrMagnitude();
            var maxMagSq = maxMagnitude * maxMagnitude;
            if (magSq > maxMagSq) {
                var ratio = Math.sqrt(maxMagSq / magSq);
                vec.x *= ratio;
                vec.y *= ratio;
            }
            return vec;
        },
        pickRandomFrom: function pickRandomFrom(arr) {
            return arr[Utils.getRandom(arr.length)];
        },
        isCounterClockwise: function isCounterClockwise(points) {
            // https://stackoverflow.com/a/14506549
            var area = 0;
            for (var i = 0, j = 0; i < points.length; i++) {
                j = (i + 1) % points.length;
                area += points[i].x * points[j].y;
                area -= points[j].x * points[i].y;
            }
            return area / 2 > 0;
        },
        round: function round(n, nearest) {
            return Math.round(n / nearest) * nearest;
        },
        approxEqual: function approxEqual(a, b, range) {
            return Math.abs(a - b) <= range;
        },
        approxEqualVec: function approxEqualVec(a, b, range) {
            return Math.abs(a.x - b.x) <= range && Math.abs(a.y - b.y) <= range;
        },
        forEachEdge: function forEachEdge(vertices, fn) {
            for (var i = 0; i < vertices.length - 1; i++) {
                fn(vertices[i], vertices[i + 1]);
            }
            fn(vertices[i], vertices[0]);
        },
        forEachCorner: function forEachCorner(vertices, fn) {
            if (vertices.length < 3) {
                return;
            }
            fn(vertices[vertices.length - 1], vertices[0], vertices[1]);
            for (var i = 1; i < vertices.length - 1; i++) {
                fn(vertices[i - 1], vertices[i], vertices[i + 1]);
            }
            fn(vertices[i - 1], vertices[i], vertices[0]);
        },
        pointOnEdge: function pointOnEdge(p, a, b) {
            var px = p.x - a.x;
            var py = p.y - a.y;
            var dx = b.x - a.x;
            var dy = b.y - a.y;
            var cross = px * dy + py * dx;
            if (Math.abs(cross) > Utils.EPSILON) {
                return false;
            }
            if (Math.abs(dx) >= Math.abs(dy)) {
                return dx > 0 ? a.x <= p.x && p.x <= b.x : b.x <= p.x && p.x <= a.x;
            } else {
                return dy > 0 ? a.y <= p.y && p.y <= b.y : b.y <= p.y && p.y <= a.y;
            }
        },
        closestPointOnEdge: function closestPointOnEdge(p, a, b) {
            var ap = p.subtract(a);
            var ab = b.subtract(a);
            var sqrMag = ab.sqrMagnitude();
            var dot = ap.dotProduct(ab);
            var t = dot / sqrMag;
            return new Vector2(a.x + ab.x * t, a.y + ab.y * t);
        },
        flatten: function flatten(arr) {
            var res = [];
            for (var i = 0; i < arr.length; i++) {
                var subarr = arr[i];
                for (var j = 0; j < subarr.length; j++) {
                    res.push(subarr[j]);
                }
            }
            return res;
        },


        /**
         * Check if three points (each of the form [x,y]) are collinear.
         * https://github.com/schteppe/poly-decomp.js
         */
        isCollinear: function () {
            var tmpPoint1 = [],
                tmpPoint2 = [];
            var triangleArea = function triangleArea(a, b, c) {
                return (b[0] - a[0]) * (c[1] - a[1]) - (c[0] - a[0]) * (b[1] - a[1]);
            };
            return function (a, b, c, thresholdAngle) {
                if (!thresholdAngle) {
                    return triangleArea(a, b, c) === 0;
                } else {
                    var ab = tmpPoint1;
                    var bc = tmpPoint2;
                    ab[0] = b[0] - a[0];
                    ab[1] = b[1] - a[1];
                    bc[0] = c[0] - b[0];
                    bc[1] = c[1] - b[1];
                    var dot = ab[0] * bc[0] + ab[1] * bc[1];
                    var magA = Math.sqrt(ab[0] * ab[0] + ab[1] * ab[1]);
                    var magB = Math.sqrt(bc[0] * bc[0] + bc[1] * bc[1]);
                    var angle = Math.acos(dot / (magA * magB));
                    return angle < thresholdAngle;
                }
            };
        }(),

        triangulateBadPolys: function triangulateBadPolys(polys) {
            var earcut = window.earcut;

            // identify any polygons which have collinear points, because they can cause janky behaviour in Box2D
            var badPolys = [];
            polys.forEach(function (poly, index) {
                var a = poly[poly.length - 2];
                var b = poly[poly.length - 1];
                var c = poly[0];
                for (var i = 1; i <= poly.length; i++) {
                    if (Utils.isCollinear(a, b, c)) {
                        badPolys.push(index);
                        break;
                    }
                    a = b;
                    b = c;
                    c = poly[i];
                }
            });

            // triangulate and remove the offenders
            for (var i = badPolys.length - 1; i >= 0; i--) {
                var index = badPolys[i];
                var poly = polys[index];
                polys.splice(index, 1);

                var triangles = earcut(Utils.flatten(poly));
                for (var j = 0; j < triangles.length; j += 3) {
                    polys.push([poly[triangles[j]], poly[triangles[j + 1]], poly[triangles[j + 2]]]);
                }
            }
        },
        newTexture: function newTexture(imageName) {
            var scaleMode = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : window.PIXI.SCALE_MODES.LINEAR;

            var PIXI = window.PIXI;
            var packedImage = Bento.assets.getImage(imageName);
            if (!packedImage) {
                Utils.log('Could not find image ' + imageName);
                return null;
            }
            var image = packedImage.image;
            if (!image.texture) {
                image.texture = new PIXI.BaseTexture(image, scaleMode);
                image.frame = new PIXI.Texture(image.texture, new PIXI.Rectangle(packedImage.x, packedImage.y, packedImage.width, packedImage.height));
                // image.frame._updateUvs();
            }
            var res = new PIXI.Texture(image.texture, new PIXI.Rectangle(packedImage.x, packedImage.y, packedImage.width, packedImage.height));
            res._updateUvs();
            return res;
        },
        getPolygonBounds: function getPolygonBounds(points) {
            var x1 = Infinity,
                y1 = Infinity;
            var x2 = -Infinity,
                y2 = -Infinity;
            points.forEach(function (v) {
                if (v.x < x1) x1 = v.x;
                if (v.x > x2) x2 = v.x;
                if (v.y < y1) y1 = v.y;
                if (v.y > y2) y2 = v.y;
            });
            if (!isFinite(x1)) x1 = 0;
            if (!isFinite(y1)) y1 = 0;
            if (!isFinite(x2)) x2 = 1;
            if (!isFinite(y2)) y2 = 1;
            return new Rectangle(x1, y1, x2 - x1, y2 - y1);
        },
        getEdgeKey: function getEdgeKey(v1, v2) {
            // sort v1,v2 so key will always be the same regardless of direction
            if (v2.y < v1.y) {
                var tmp = v1;
                v1 = v2;
                v2 = tmp;
            }
            if (v2.x < v1.x) {
                var _tmp = v1;
                v1 = v2;
                v2 = _tmp;
            }
            // key should be reasonably short and not contain floating point inaccuracies
            var x1 = Utils.round(v1.x, 0.01);
            var y1 = Utils.round(v1.y, 0.01);
            var x2 = Utils.round(v2.x, 0.01);
            var y2 = Utils.round(v2.y, 0.01);

            return x1 + ',' + y1 + ':' + x2 + ',' + y2;
        },
        getParent: function getParent(entity, parentName) {
            var parent = entity.parent;
            if (!parent) {
                return null;
            }
            while (parent.name !== parentName) {
                if (parent.parent) {
                    parent = parent.parent;
                } else {
                    return null;
                }
            }
            return parent;
        },
        didScroll: function didScroll(entity) {
            var parent = Utils.getParent(entity, 'scrollingList');
            return parent !== null && parent.didMove(4);
        },
        didScrollOrMissedArea: function didScrollOrMissedArea(entity, mouseEvent) {
            var parent = Utils.getParent(entity, 'scrollingList');
            if (parent == null) {
                return false;
            }
            if (parent.didMove(4)) {
                return true;
            }
            var area = parent.area;
            return area && !area.hasPosition(parent.toLocalPosition(mouseEvent.worldPosition));
        },
        sequenceTweens: function sequenceTweens(settingsArr) {
            if (settingsArr.length === undefined) {
                // allow varargs style call
                Utils.sequenceTweens(arguments);
                return;
            }
            var i,
                settings,
                prevSettings = {};
            for (i = 0; i < settingsArr.length; i++) {
                settings = settingsArr[i];
                settings.delay = (settings.delay || 0) + (prevSettings.delay || 0) + (prevSettings.in || 0);
                settings.applyOnDelay = false;
                if (!settings.onUpdate) {
                    settings.onUpdate = prevSettings.onUpdate;
                }
                if (settings.from === undefined) {
                    settings.from = prevSettings.to;
                }
                new Tween(settings);
                prevSettings = settings;
            }
        },
        log: function log(msg) {
            console.error(msg);
            EventSystem.fire('GameAnalytics-addErrorEvent', {
                severity: 2, // Info
                message: msg
            });
        },
        currentScreenName: function currentScreenName() {
            var screen = Bento.screens.getCurrentScreen();
            if (!screen) {
                return 'noscreen';
            }
            var parts = screen.name.split('/');
            return parts[1] || parts[0];
        },
        getPlatform: function getPlatform() {
            if (Utils.isNativeAndroid()) {
                return 'android';
            } else if (Utils.isNativeIos()) {
                return 'ios';
                // } else if (Utils.isFB()) {
                // return 'fb';
            } else {
                return 'web';
            }
        },
        getHashParameter: function getHashParameter(key, fallback) {
            var url = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : window.location.hash;

            var match = RegExp('[#&]' + key + '=([^&]*)').exec(url);
            var result = match && decodeURIComponent(match[1].replace(/\+/g, ' '));

            if (result !== null) {
                return result;
            } else if (Utils.isDefined(fallback)) {
                return fallback;
            } else {
                return null;
            }
        },
        setupHDSprite: function setupHDSprite(obj, settings) {
            // potentially useful except it's common for an entity to have 2 hdSprites (e.g. one as a drop shadow)
            for (var i = 0; i < obj.components.length; i++) {
                var c = obj.components[i];
                if (c && c.components) {
                    for (var j = 0; j < c.components.length; j++) {
                        // hdsprite always has a hdSpriteBehavior component
                        if (c.components[j].name === 'hdSpriteBehavior') {
                            // inner sprite always has the same name as the hdsprite itself
                            c.getComponent(c.name, function (sprite) {
                                sprite.setup(settings);
                            });
                            return;
                        }
                    }
                }
            }
        },
        refilter: function refilter(fixture) {
            if (!fixture.m_body) {
                return;
            }
            var edge = fixture.m_body.GetContactList();
            while (edge) {
                var contact = edge.contact;
                var fixtureA = contact.GetFixtureA();
                var fixtureB = contact.GetFixtureB();
                if (fixtureA === fixture || fixtureB === fixture) {
                    contact.FlagForFiltering();
                }
                edge = edge.next;
            }
            var world = fixture.m_body.GetWorld();
            if (!world) {
                return;
            }
            var broadPhase = world.m_contactManager.m_broadPhase;
            if (fixture.m_proxy) {
                broadPhase.BufferMove(fixture.m_proxy.proxyId);
            }
        },
        stringHash: function stringHash(str) {
            var hash = 0,
                i,
                chr;
            if (str.length === 0) return hash;
            for (i = 0; i < str.length; i++) {
                chr = str.charCodeAt(i);
                hash = (hash << 5) - hash + chr;
                hash |= 0; // Convert to 32bit integer
            }
            return hash;
        },
        zeroPad: function zeroPad(num, size) {
            var s = num + "";
            while (s.length < size) {
                s = "0" + s;
            }return s;
        }
    });

    return Utils;
});